﻿using System;
using System.Linq.Expressions;
using System.Reflection;

namespace Nito.AsyncEx.Internal.PlatformEnlightenment
{
    partial class DefaultEnlightenmentProvider
    {
        /// <summary>
        /// The default exception enlightenment, which will use <c>ExceptionDispatchInfo</c> if possible, falling back on <c>Exception.PrepForRemoting</c>, with a final fallback on <see cref="Exception.Data"/>.
        /// </summary>
        public sealed class ExceptionEnlightenment : IExceptionEnlightenment
        {
            /// <summary>
            /// A delegate that will call <c>ExceptionDispatchInfo.Capture</c> followed by <c>ExceptionDispatchInfo.Throw</c>, or <c>null</c> if the <c>ExceptionDispatchInfo</c> type does not exist.
            /// </summary>
            private static readonly Action<Exception> CaptureAndThrow;

            /// <summary>
            /// A delegate that will call <c>Exception.PrepForRemoting</c>, or <c>null</c> if the method does not exist. This member is always <c>null</c> if <see cref="CaptureAndThrow"/> is non-<c>null</c>.
            /// </summary>
            private static readonly Action<Exception> PrepForRemoting;

            /// <summary>
            /// Attempts to look up a method for a type, handling vexing exceptions.
            /// </summary>
            /// <param name="type">The type on which to look up the method.</param>
            /// <param name="name">The method to look up.</param>
            /// <param name="flags">The binding flags used when searching for the method.</param>
            private static MethodInfo TryGetMethod(Type type, string name, BindingFlags flags)
            {
                try
                {
                    return type.GetMethod(name, flags);
                }
                catch (AmbiguousMatchException)
                {
                    // vexing exception
                    return null;
                }
            }

            /// <summary>
            /// Examines the current runtime and initializes the static delegates appropriately.
            /// </summary>
            static ExceptionEnlightenment()
            {
                var exceptionType = typeof(Exception);
                var exceptionDispatchInfoType = Type.GetType("System.Runtime.ExceptionServices.ExceptionDispatchInfo");
                if (exceptionDispatchInfoType != null)
                {
                    try
                    {
                        var parameter = Expression.Parameter(exceptionType, "exception");
                        var captureCall = Expression.Call(exceptionDispatchInfoType, "Capture", null, parameter);
                        var throwCall = Expression.Call(captureCall, "Throw", null);
                        CaptureAndThrow = Expression.Lambda<Action<Exception>>(throwCall, parameter).Compile();
                    }
                    catch (InvalidOperationException)
                    {
                        // vexing exception (Expression.Call)
                    }
                    catch (ArgumentException)
                    {
                        // vexing exception (Expression.Lambda)
                    }
                }

                if (CaptureAndThrow == null)
                {
                    var prepForRemotingMethod = TryGetMethod(exceptionType, "PrepForRemoting", BindingFlags.Instance | BindingFlags.NonPublic);
                    if (prepForRemotingMethod != null)
                    {
                        try
                        {
                            var parameter = Expression.Parameter(exceptionType, "exception");
                            var call = Expression.Call(parameter, prepForRemotingMethod);
                            PrepForRemoting = Expression.Lambda<Action<Exception>>(call, parameter).Compile();
                        }
                        catch (ArgumentException)
                        {
                            // vexing exception (Expression.Call, Expression.Lambda)
                        }
                    }
                }
            }

            /// <summary>
            /// Attempts to add the original stack trace to the <see cref="Exception.Data"/> collection.
            /// </summary>
            /// <param name="exception">The exception. May not be <c>null</c>.</param>
            /// <returns><c>true</c> if the stack trace was successfully saved; <c>false</c> otherwise.</returns>
            private static void TryAddStackTrace(Exception exception)
            {
                try
                {
                    exception.Data.Add("Original stack trace", exception.StackTrace);
                }
                catch (ArgumentException)
                {
                    // Vexing exception
                }
                catch (NotSupportedException)
                {
                    // Vexing exception
                }
            }

            Exception IExceptionEnlightenment.PrepareForRethrow(Exception exception)
            {
                if (CaptureAndThrow != null)
                {
                    CaptureAndThrow(exception);
                }
                else if (PrepForRemoting != null)
                {
                    PrepForRemoting(exception);
                }
                else
                {
                    TryAddStackTrace(exception);
                }

                return exception;
            }
        }
    }
}
