﻿using System;
using System.Reflection;
using System.Threading;

namespace Nito.AsyncEx.Internal.PlatformEnlightenment
{
    /// <summary>
    /// Provides static members to access enlightenments.
    /// </summary>
    public static class Enlightenment
    {
        /// <summary>
        /// Cached instance of the platform enlightenment provider. This is an instance of the default enlightenment provider if the platform couldn't be found.
        /// </summary>
        private static IEnlightenmentProvider _platform;

        /// <summary>
        /// Cached instance of the exception enlightenment.
        /// </summary>
        private static IExceptionEnlightenment _exception;

        /// <summary>
        /// Cached instance of the async enlightenment.
        /// </summary>
        private static IAsyncEnlightenment _async;

        /// <summary>
        /// Cached instance of the synchronization context enlightenment.
        /// </summary>
        private static ISynchronizationContextEnlightenment _synchronizationContext;

        /// <summary>
        /// Cached instance of the trace enlightenment.
        /// </summary>
        private static ITraceEnlightenment _trace;

        /// <summary>
        /// Cached instance of the concurrent collections enlightenment.
        /// </summary>
        private static IConcurrentCollectionsEnlightenment _concurrentCollections;

        /// <summary>
        /// Cached instance of the lazy enlightenment.
        /// </summary>
        private static ILazyEnlightenment _lazy;

        /// <summary>
        /// Loads the <c>PlatformEnlightenmentProvider</c> if it can be found; otherwise, returns an instance of <see cref="DefaultEnlightenmentProvider"/>.
        /// </summary>
        private static IEnlightenmentProvider CreateProvider()
        {
            var enlightenmentAssemblyName = new AssemblyName(ReflectionShim.GetAssembly(typeof(IEnlightenmentProvider)).FullName)
            {
                Name = "Nito.AsyncEx.Enlightenment",
            };
            var enlightenmentProviderType = Type.GetType("Nito.AsyncEx.Internal.PlatformEnlightenment.EnlightenmentProvider, " + enlightenmentAssemblyName.FullName, false);
            if (enlightenmentProviderType == null)
                return new DefaultEnlightenmentProvider();
            return (IEnlightenmentProvider)Activator.CreateInstance(enlightenmentProviderType);
        }

        /// <summary>
        /// Returns the platform enlightenment provider, if it could be found; otherwise, returns the default enlightenment provider.
        /// </summary>
        public static IEnlightenmentProvider Platform
        {
            get
            {
                if (_platform == null)
                    Interlocked.CompareExchange(ref _platform, CreateProvider(), null);
                return _platform;
            }
        }

        /// <summary>
        /// Returns the exception enlightenment.
        /// </summary>
        public static IExceptionEnlightenment Exception
        {
            get
            {
                if (_exception == null)
                    Interlocked.CompareExchange(ref _exception, Platform.CreateEnlightenment<IExceptionEnlightenment>(), null);
                return _exception;
            }
        }

        /// <summary>
        /// Returns the async enlightenment.
        /// </summary>
        public static IAsyncEnlightenment Async
        {
            get
            {
                if (_async == null)
                    Interlocked.CompareExchange(ref _async, Platform.CreateEnlightenment<IAsyncEnlightenment>(), null);
                return _async;
            }
        }

        /// <summary>
        /// Returns the synchronization context enlightenment.
        /// </summary>
        public static ISynchronizationContextEnlightenment SynchronizationContext
        {
            get
            {
                if (_synchronizationContext == null)
                    Interlocked.CompareExchange(ref _synchronizationContext, Platform.CreateEnlightenment<ISynchronizationContextEnlightenment>(), null);
                return _synchronizationContext;
            }
        }

        /// <summary>
        /// Returns the trace enlightenment.
        /// </summary>
        public static ITraceEnlightenment Trace
        {
            get
            {
                if (_trace == null)
                    Interlocked.CompareExchange(ref _trace, Platform.CreateEnlightenment<ITraceEnlightenment>(), null);
                return _trace;
            }
        }

        /// <summary>
        /// Returns the concurrent collections enlightenment.
        /// </summary>
        public static IConcurrentCollectionsEnlightenment ConcurrentCollections
        {
            get
            {
                if (_concurrentCollections == null)
                    Interlocked.CompareExchange(ref _concurrentCollections, Platform.CreateEnlightenment<IConcurrentCollectionsEnlightenment>(), null);
                return _concurrentCollections;
            }
        }

        /// <summary>
        /// Returns the lazy enlightenment.
        /// </summary>
        public static ILazyEnlightenment Lazy
        {
            get
            {
                if (_lazy == null)
                    Interlocked.CompareExchange(ref _lazy, Platform.CreateEnlightenment<ILazyEnlightenment>(), null);
                return _lazy;
            }
        }
    }
}
