﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Collections.Generic;
using System.Threading.Tasks.Dataflow;
using System.Threading;
using System.Diagnostics.CodeAnalysis;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class FuncBlockUnitTests_NET40
    {
        private static async Task<IList<T>> GetBlockOutputsAsync<T>(ISourceBlock<T> block)
        {
            var ret = new List<T>();
            var actionBlock = new ActionBlock<T>(item =>
            {
                ret.Add(item);
            });
            block.LinkTo(actionBlock);
            var __ = block.Completion.ContinueWith(_ => actionBlock.Complete());
            await actionBlock.Completion;
            return ret;
        }

        [TestMethod]
        public void FuncBlock_FuncProducesItemsAndCompletes_ProducesItemsAndCompletes()
        {
            AsyncContext.Run(async () =>
            {
                var block = new FuncBlock<int>(async send =>
                {
                    await send(13);
                    await send(17);
                    await send(19);
                });

                var results = await GetBlockOutputsAsync(block);
                await block.Completion;
                Assert.IsTrue(results.SequenceEqual(new[] { 13, 17, 19 }));
            });
        }

        [TestMethod]
        public void FuncBlock_FuncFaults_Faults()
        {
            AsyncContext.Run(async () =>
            {
                var block = new FuncBlock<int>(async send =>
                {
                    await send(13);
                    throw new NotImplementedException();
                });

                await GetBlockOutputsAsync(block);
                await AssertEx.ThrowsExceptionAsync<NotImplementedException>(block.Completion, allowDerivedTypes: false);
            });
        }

        [TestMethod]
        public void FuncBlock_FuncCancels_Completes()
        {
            AsyncContext.Run(async () =>
            {
                var cts = new CancellationTokenSource();
                cts.Cancel();
                var block = new FuncBlock<int>(async send =>
                {
                    await send(13);
                    cts.Token.ThrowIfCancellationRequested();
                });

                var results = await GetBlockOutputsAsync(block);
                await block.Completion;
                Assert.IsTrue(results.SequenceEqual(new[] { 13 }));
            });
        }

        [TestMethod]
        public void FuncBlock_Canceled_Cancels()
        {
            AsyncContext.Run(async () =>
            {
                var cts = new CancellationTokenSource();
                var block = new FuncBlock<int>(async send =>
                {
                    while (true)
                    {
                        await send(13);
                        await TaskEx.Delay(100);
                    }
                },
                    new DataflowBlockOptions
                    {
                        CancellationToken = cts.Token,
                    });

                var resultTask = GetBlockOutputsAsync(block);
                cts.Cancel();
                var results = await resultTask;
                await AssertEx.CompletesCanceledAsync(block.Completion);
            });
        }

        [TestMethod]
        public void FuncBlock_Canceled_CancelsFunc()
        {
            AsyncContext.Run(async () =>
            {
                var tcs = new TaskCompletionSource();
                var cts = new CancellationTokenSource();
                var block = new FuncBlock<int>(async send =>
                {
                    try
                    {
                        while (true)
                        {
                            await send(13);
                            await TaskEx.Delay(100);
                        }
                    }
                    catch (OperationCanceledException)
                    {
                        tcs.SetResult();
                    }
                },
                    new DataflowBlockOptions
                    {
                        CancellationToken = cts.Token,
                    });

                var resultTask = GetBlockOutputsAsync(block);
                cts.Cancel();
                await resultTask;
                await tcs.Task;
            });
        }

        [TestMethod]
        public void FuncBlock_Completed_Completes()
        {
            AsyncContext.Run(async () =>
            {
                var block = new FuncBlock<int>(async send =>
                {
                    while (true)
                    {
                        await send(13);
                        await TaskEx.Delay(100);
                    }
                });

                var resultTask = GetBlockOutputsAsync(block);
                block.Complete();
                await resultTask;
                await block.Completion;
            });
        }

        [TestMethod]
        public void FuncBlock_Completed_CancelsFunc()
        {
            AsyncContext.Run(async () =>
            {
                var tcs = new TaskCompletionSource();
                var block = new FuncBlock<int>(async send =>
                {
                    try
                    {
                        while (true)
                        {
                            await send(13);
                            await TaskEx.Delay(100);
                        }
                    }
                    catch (OperationCanceledException)
                    {
                        tcs.SetResult();
                    }
                });

                var resultTask = GetBlockOutputsAsync(block);
                block.Complete();
                await resultTask;
                await block.Completion;
                await tcs.Task;
            });
        }

        [TestMethod]
        public void FuncBlock_Faulted_Faults()
        {
            AsyncContext.Run(async () =>
            {
                var block = new FuncBlock<int>(async send =>
                {
                    while (true)
                    {
                        await send(13);
                        await TaskEx.Delay(100);
                    }
                });

                ((IDataflowBlock)block).Fault(new NotImplementedException());
                await AssertEx.ThrowsExceptionAsync<NotImplementedException>(block.Completion, allowDerivedTypes: false);
            });
        }

        [TestMethod]
        public void FuncBlock_Faulted_CancelsFunc()
        {
            AsyncContext.Run(async () =>
            {
                var tcs = new TaskCompletionSource();
                var block = new FuncBlock<int>(async send =>
                {
                    try
                    {
                        while (true)
                        {
                            await send(13);
                            await TaskEx.Delay(100);
                        }
                    }
                    catch (OperationCanceledException)
                    {
                        tcs.SetResult();
                    }
                });

                ((IDataflowBlock)block).Fault(new NotImplementedException());
                await tcs.Task;
            });
        }
    }
}
