﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncCountdownEventUnitTests_NET40_NE
    {
        [TestMethod]
        public void WaitAsync_Unset_IsNotCompleted()
        {
            AsyncContext.Run(async () =>
            {
                var ce = new AsyncCountdownEvent(1);
                var task = ce.WaitAsync();
                Assert.AreEqual(1, ce.CurrentCount);
                await AssertEx.NeverCompletesAsync(task);
            });
        }

        [TestMethod]
        public void WaitAsync_AfterCountingDown_IsCompleted()
        {
            var ce = new AsyncCountdownEvent(2);
            Assert.AreEqual(2, ce.CurrentCount);
            ce.Signal();
            var task = ce.WaitAsync();
            Assert.AreEqual(1, ce.CurrentCount);
            Assert.IsFalse(task.IsCompleted);
            ce.Signal();
            Assert.AreEqual(0, ce.CurrentCount);
            Assert.IsTrue(task.IsCompleted);
        }

        [TestMethod]
        public void AddCount_IncrementsCount()
        {
            var ce = new AsyncCountdownEvent(1);
            var task = ce.WaitAsync();
            Assert.AreEqual(1, ce.CurrentCount);
            Assert.IsFalse(task.IsCompleted);
            ce.AddCount();
            Assert.AreEqual(2, ce.CurrentCount);
            Assert.IsFalse(task.IsCompleted);
            ce.Signal();
            Assert.AreEqual(1, ce.CurrentCount);
            Assert.IsFalse(task.IsCompleted);
            ce.Signal();
            Assert.AreEqual(0, ce.CurrentCount);
            Assert.IsTrue(task.IsCompleted);
        }

        [TestMethod]
        public void Signal_AfterSet_ThrowsException()
        {
            var ce = new AsyncCountdownEvent(1);
            ce.Signal();
            AssertEx.ThrowsException<InvalidOperationException>(() => ce.Signal());
        }

        [TestMethod]
        public void TrySignal_AfterSet_ReturnsFalse()
        {
            var ce = new AsyncCountdownEvent(1);
            ce.Signal();
            var result = ce.TrySignal();
            Assert.IsFalse(result);
        }

        [TestMethod]
        public void AddCount_AfterSet_ThrowsException()
        {
            var ce = new AsyncCountdownEvent(1);
            ce.Signal();
            AssertEx.ThrowsException<InvalidOperationException>(() => ce.AddCount());
        }

        [TestMethod]
        public void TryAddCount_AfterSet_ReturnsFalse()
        {
            var ce = new AsyncCountdownEvent(1);
            ce.Signal();
            var result = ce.TryAddCount();
            Assert.IsFalse(result);
        }

        [TestMethod]
        public void AddCount_Overflow_ThrowsException()
        {
            var ce = new AsyncCountdownEvent(int.MaxValue);
            AssertEx.ThrowsException<InvalidOperationException>(() => ce.AddCount());
        }

        [TestMethod]
        public void Id_EqualsTaskId()
        {
            var ce = new AsyncCountdownEvent(1);
            var task = ce.WaitAsync();
            Assert.AreEqual(task.Id, ce.Id);
        }
    }
}
