﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncMonitorUnitTests_NET45_NE
    {
        [TestMethod]
        public async Task Unlocked_PermitsLock()
        {
            var monitor = new AsyncMonitor();
            
            var task = monitor.EnterAsync();
            await task;
        }

        [TestMethod]
        public async Task Locked_PreventsLockUntilUnlocked()
        {
            var monitor = new AsyncMonitor();
            var task1HasLock = new TaskCompletionSource();
            var task1Continue = new TaskCompletionSource();
            Task<IDisposable> initialLockTask = null;

            var task1 = Task.Run(async () =>
            {
                initialLockTask = monitor.EnterAsync();
                using (await initialLockTask)
                {
                    task1HasLock.SetResult();
                    await task1Continue.Task;
                }
            });
            await task1HasLock.Task;

            var lockTask = monitor.EnterAsync();
            Assert.IsFalse(lockTask.IsCompleted);
            task1Continue.SetResult();
            await lockTask;
        }

        [TestMethod]
        public async Task Pulse_ReleasesOneWaiter()
        {
            var monitor = new AsyncMonitor();
            int completed = 0;
            var task1Ready = new TaskCompletionSource();
            var task2Ready = new TaskCompletionSource();
            Task<IDisposable> lockTask1 = null;
            Task waitTask1 = null;
            var task1 = Task.Run(async () =>
            {
                lockTask1 = monitor.EnterAsync();
                using (await lockTask1)
                {
                    waitTask1 = monitor.WaitAsync();
                    task1Ready.SetResult();
                    await waitTask1;
                    Interlocked.Increment(ref completed);
                }
            });
            await task1Ready.Task;
            Task<IDisposable> lockTask2 = null;
            Task waitTask2 = null;
            var task2 = Task.Run(async () =>
            {
                lockTask2 = monitor.EnterAsync();
                using (await lockTask2)
                {
                    waitTask2 = monitor.WaitAsync();
                    task2Ready.SetResult();
                    await waitTask2;
                    Interlocked.Increment(ref completed);
                }
            });
            await task2Ready.Task;

            Task<IDisposable> lockTask3 = monitor.EnterAsync();
            using (await lockTask3)
            {
                monitor.Pulse();
            }
            await Task.WhenAny(task1, task2);
            var result = Interlocked.CompareExchange(ref completed, 0, 0);

            Assert.AreEqual(1, result);
        }

        [TestMethod]
        public async Task PulseAll_ReleasesAllWaiters()
        {
            var monitor = new AsyncMonitor();
            int completed = 0;
            var task1Ready = new TaskCompletionSource();
            var task2Ready = new TaskCompletionSource();
            Task<IDisposable> lockTask1 = null;
            Task waitTask1 = null;
            var task1 = Task.Run(async () =>
            {
                lockTask1 = monitor.EnterAsync();
                using (await lockTask1)
                {
                    waitTask1 = monitor.WaitAsync();
                    task1Ready.SetResult();
                    await waitTask1;
                    Interlocked.Increment(ref completed);
                }
            });
            await task1Ready.Task;
            Task<IDisposable> lockTask2 = null;
            Task waitTask2 = null;
            var task2 = Task.Run(async () =>
            {
                lockTask2 = monitor.EnterAsync();
                using (await lockTask2)
                {
                    waitTask2 = monitor.WaitAsync();
                    task2Ready.SetResult();
                    await waitTask2;
                    Interlocked.Increment(ref completed);
                }
            });
            await task2Ready.Task;

            var lockTask3 = monitor.EnterAsync();
            using (await lockTask3)
            {
                monitor.PulseAll();
            }
            await Task.WhenAll(task1, task2);
            var result = Interlocked.CompareExchange(ref completed, 0, 0);
            
            Assert.AreEqual(2, result);
        }

        [TestMethod]
        public void Id_IsNotZero()
        {
            var monitor = new AsyncMonitor();
            Assert.AreNotEqual(0, monitor.Id);
        }
    }
}
