﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncProducerConsumerQueueUnitTests_NET45_NE
    {
        [TestMethod]
        public void ConstructorWithZeroMaxCount_Throws()
        {
            AssertEx.ThrowsException<ArgumentOutOfRangeException>(() => new AsyncProducerConsumerQueue<int>(0));
        }

        [TestMethod]
        public void ConstructorWithZeroMaxCountAndCollection_Throws()
        {
            AssertEx.ThrowsException<ArgumentOutOfRangeException>(() => new AsyncProducerConsumerQueue<int>(new int[0], 0));
        }

        [TestMethod]
        public void ConstructorWithMaxCountSmallerThanCollectionCount_Throws()
        {
            AssertEx.ThrowsException<ArgumentException>(() => new AsyncProducerConsumerQueue<int>(new[] { 3, 5 }, 1));
        }

        [TestMethod]
        public async Task ConstructorWithCollection_AddsItems()
        {
            var queue = new AsyncProducerConsumerQueue<int>(new[] { 3, 5, 7 });

            var result1 = await queue.DequeueAsync();
            var result2 = await queue.DequeueAsync();
            var result3 = await queue.DequeueAsync();

            Assert.AreEqual(3, result1);
            Assert.AreEqual(5, result2);
            Assert.AreEqual(7, result3);
        }

        [TestMethod]
        public async Task EnqueueAsync_SpaceAvailable_EnqueuesItem()
        {
            var queue = new AsyncProducerConsumerQueue<int>();

            await queue.EnqueueAsync(3);
            var result = await queue.DequeueAsync();

            Assert.AreEqual(3, result);
        }

        [TestMethod]
        public async Task TryEnqueueAsync_SpaceAvailable_EnqueuesItem()
        {
            var queue = new AsyncProducerConsumerQueue<int>();

            var result = await queue.TryEnqueueAsync(3);
            var dequeueResult = await queue.DequeueAsync();

            Assert.IsTrue(result);
            Assert.AreEqual(3, dequeueResult);
        }

        [TestMethod]
        public async Task EnqueueToAnyAsync_SpaceAvailable_EnqueuesItem()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>(new[] { 3 }, 1);
            var queue2 = new AsyncProducerConsumerQueue<int>();
            var queues = new[] { queue1, queue2 };

            var result = await queues.EnqueueToAnyAsync(13);
            var dequeueResult = await queue2.DequeueAsync();

            Assert.AreSame(queue2, result);
            Assert.AreEqual(13, dequeueResult);
        }

        [TestMethod]
        public async Task TryEnqueueToAnyAsync_SpaceAvailable_EnqueuesItem()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            var queue2 = new AsyncProducerConsumerQueue<int>(new[] { 3 }, 1);
            var queues = new[] { queue1, queue2 };

            var result = await queues.TryEnqueueToAnyAsync(13);
            var dequeueResult = await queue1.DequeueAsync();

            Assert.AreSame(queue1, result);
            Assert.AreEqual(13, dequeueResult);
        }

        [TestMethod]
        public async Task TryEnqueueAsync_CompleteAdding_ReturnsFailed()
        {
            var queue = new AsyncProducerConsumerQueue<int>();
            await queue.CompleteAddingAsync();

            var result = await queue.TryEnqueueAsync(3);

            Assert.IsFalse(result);
        }

        [TestMethod]
        public async Task EnqueueAsync_CompleteAdding_ThrowsException()
        {
            var queue = new AsyncProducerConsumerQueue<int>();
            await queue.CompleteAddingAsync();

            await AssertEx.ThrowsExceptionAsync<InvalidOperationException>(() => queue.EnqueueAsync(3));
        }

        [TestMethod]
        public async Task TryEnqueueToAnyAsync_CompleteAdding_ReturnsFailed()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            await queue1.CompleteAddingAsync();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            await queue2.CompleteAddingAsync();
            var queues = new[] { queue1, queue2 };

            var result = await queues.TryEnqueueToAnyAsync(3);

            Assert.IsNull(result);
        }

        [TestMethod]
        public async Task EnqueueToAnyAsync_CompleteAdding_ThrowsException()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            await queue1.CompleteAddingAsync();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            await queue2.CompleteAddingAsync();
            var queues = new[] { queue1, queue2 };

            await AssertEx.ThrowsExceptionAsync<InvalidOperationException>(() => queues.EnqueueToAnyAsync(3));
        }

        [TestMethod]
        public async Task DequeueAsync_EmptyAndComplete_ThrowsException()
        {
            var queue = new AsyncProducerConsumerQueue<int>();
            await queue.CompleteAddingAsync();

            await AssertEx.ThrowsExceptionAsync<InvalidOperationException>(() => queue.DequeueAsync());
        }

        [TestMethod]
        public async Task TryDequeueAsync_EmptyAndComplete_ReturnsFailed()
        {
            var queue = new AsyncProducerConsumerQueue<int>();
            await queue.CompleteAddingAsync();

            var result = await queue.TryDequeueAsync();

            Assert.IsFalse(result.Success);
        }

        [TestMethod]
        public async Task DequeueFromAnyAsync_EmptyAndComplete_ThrowsException()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            await queue1.CompleteAddingAsync();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            await queue2.CompleteAddingAsync();
            var queues = new[] { queue1, queue2 };

            await AssertEx.ThrowsExceptionAsync<InvalidOperationException>(() => queues.DequeueFromAnyAsync());
        }

        [TestMethod]
        public async Task TryDequeueFromAnyAsync_EmptyAndComplete_ReturnsFailed()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            await queue1.CompleteAddingAsync();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            await queue2.CompleteAddingAsync();
            var queues = new[] { queue1, queue2 };

            var result = await queues.TryDequeueFromAnyAsync();

            Assert.IsFalse(result.Success);
        }

        [TestMethod]
        public async Task DequeueAsync_Empty_DoesNotComplete()
        {
            var queue = new AsyncProducerConsumerQueue<int>();

            var task = queue.DequeueAsync();

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task TryDequeueAsync_Empty_DoesNotComplete()
        {
            var queue = new AsyncProducerConsumerQueue<int>();

            var task = queue.TryDequeueAsync();

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task DequeueFromAnyAsync_Empty_DoesNotComplete()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            var queues = new[] { queue1, queue2 };

            var task = queues.DequeueFromAnyAsync();

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task TryDequeueFromAnyAsync_Empty_DoesNotComplete()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            var queues = new[] { queue1, queue2 };

            var task = queues.TryDequeueFromAnyAsync();

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task DequeueAsync_Empty_ItemAdded_Completes()
        {
            var queue = new AsyncProducerConsumerQueue<int>();
            var task = queue.DequeueAsync();

            await queue.EnqueueAsync(13);
            var result = await task;

            Assert.AreEqual(13, result);
        }

        [TestMethod]
        public async Task DequeueFromAnyAsync_Empty_ItemAdded_Completes()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            var queues = new[] { queue1, queue2 };
            var task = queues.DequeueFromAnyAsync();

            await queue2.EnqueueAsync(13);
            var result = await task;

            Assert.IsTrue(result.Success);
            Assert.AreSame(queue2, result.Queue);
            Assert.AreEqual(13, result.Item);
        }

        [TestMethod]
        public async Task TryDequeueAsync_Cancelled_Throws()
        {
            var queue = new AsyncProducerConsumerQueue<int>();
            var cts = new CancellationTokenSource();
            var task = queue.TryDequeueAsync(cts.Token);

            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
        }

        [TestMethod]
        public async Task DequeueAsync_Cancelled_Throws()
        {
            var queue = new AsyncProducerConsumerQueue<int>();
            var cts = new CancellationTokenSource();
            var task = queue.DequeueAsync(cts.Token);

            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
        }

        [TestMethod]
        public async Task TryDequeueFromAnyAsync_Cancelled_Throws()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            var queues = new[] { queue1, queue2 };
            var cts = new CancellationTokenSource();
            var task = queues.TryDequeueFromAnyAsync(cts.Token);

            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
        }

        [TestMethod]
        public async Task DequeueFromAnyAsync_Cancelled_Throws()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            var queues = new[] { queue1, queue2 };
            var cts = new CancellationTokenSource();
            var task = queues.DequeueFromAnyAsync(cts.Token);

            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
        }

        [TestMethod]
        public async Task EnqueueAsync_Full_DoesNotComplete()
        {
            var queue = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);

            var task = queue.EnqueueAsync(7);

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task TryEnqueueAsync_Full_DoesNotComplete()
        {
            var queue = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);

            var task = queue.TryEnqueueAsync(7);

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task EnqueueToAnyAsync_Full_DoesNotComplete()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queue2 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queues = new[] { queue1, queue2 };

            var task = queues.EnqueueToAnyAsync(7);

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task TryEnqueueToAnyAsync_Full_DoesNotComplete()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queue2 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queues = new[] { queue1, queue2 };

            var task = queues.TryEnqueueToAnyAsync(7);

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task EnqueueAsync_SpaceAvailable_Completes()
        {
            var queue = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var task = queue.EnqueueAsync(7);

            await queue.DequeueAsync();

            await task;
        }

        [TestMethod]
        public async Task TryEnqueueAsync_SpaceAvailable_Completes()
        {
            var queue = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var task = queue.TryEnqueueAsync(7);

            await queue.DequeueAsync();
            var result = await task;

            Assert.IsTrue(result);
        }

        [TestMethod]
        public async Task EnqueueToAnyAsync_SpaceAvailable_Completes()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queue2 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queues = new[] { queue1, queue2 };
            var task = queues.EnqueueToAnyAsync(7);

            await queue1.DequeueAsync();
            var result = await task;

            Assert.AreSame(queue1, result);
        }

        [TestMethod]
        public async Task TryEnqueueToAnyAsync_SpaceAvailable_Completes()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queue2 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queues = new[] { queue1, queue2 };
            var task = queues.TryEnqueueToAnyAsync(7);

            await queue2.DequeueAsync();
            var result = await task;

            Assert.AreSame(queue2, result);
        }

        [TestMethod]
        public async Task TryEnqueueAsync_Cancelled_Throws()
        {
            var queue = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var cts = new CancellationTokenSource();
            var task = queue.TryEnqueueAsync(7, cts.Token);

            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
        }

        [TestMethod]
        public async Task EnqueueAsync_Cancelled_Throws()
        {
            var queue = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var cts = new CancellationTokenSource();
            var task = queue.EnqueueAsync(7, cts.Token);

            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
        }

        [TestMethod]
        public async Task TryEnqueueToAnyAsync_Cancelled_Throws()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queue2 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queues = new[] { queue1, queue2 };
            var cts = new CancellationTokenSource();
            var task = queues.TryEnqueueToAnyAsync(7, cts.Token);

            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
        }

        [TestMethod]
        public async Task EnqueueToAnyAsync_Cancelled_Throws()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queue2 = new AsyncProducerConsumerQueue<int>(new[] { 13 }, 1);
            var queues = new[] { queue1, queue2 };
            var cts = new CancellationTokenSource();
            var task = queues.EnqueueToAnyAsync(7, cts.Token);

            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
        }

        [TestMethod]
        public async Task EnqueueToAnyAsync_MultipleHaveRoom_OnlyEnqueuesOne()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>();
            var queue2 = new AsyncProducerConsumerQueue<int>();
            var queues = new[] { queue1, queue2 };

            var result = await queues.EnqueueToAnyAsync(13);
            Task task = ((result == queue1) ? queue2 : queue1).DequeueAsync();
            await result.DequeueAsync();

            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task DequeueFromAnyAsync_MultipleHaveEntries_OnlyDequeuesOne()
        {
            var queue1 = new AsyncProducerConsumerQueue<int>(new[] { 5 });
            var queue2 = new AsyncProducerConsumerQueue<int>(new[] { 7 });
            var queues = new[] { queue1, queue2 };

            var result = await queues.DequeueFromAnyAsync();
            await ((result.Queue == queue1) ? queue2 : queue1).DequeueAsync();
        }

        [TestMethod]
        public async Task CompleteAdding_MultipleTimes_DoesNotThrow()
        {
            var queue = new AsyncProducerConsumerQueue<int>();
            await queue.CompleteAddingAsync();

            await queue.CompleteAddingAsync();
        }
    }
}
