﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Collections.Generic;
using System.Threading.Tasks.Dataflow;
using System.Threading;
using System.Diagnostics.CodeAnalysis;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class FuncBlockUnitTests_NET45_NE
    {
        private static async Task<IList<T>> GetBlockOutputsAsync<T>(ISourceBlock<T> block)
        {
            var ret = new List<T>();
            var actionBlock = new ActionBlock<T>(item =>
            {
                ret.Add(item);
            });
            block.LinkTo(actionBlock);
            var __ = block.Completion.ContinueWith(_ => actionBlock.Complete());
            await actionBlock.Completion;
            return ret;
        }

        [TestMethod]
        public async Task FuncBlock_FuncProducesItemsAndCompletes_ProducesItemsAndCompletes()
        {
            var block = new FuncBlock<int>(async send =>
            {
                await send(13);
                await send(17);
                await send(19);
            });

            var results = await GetBlockOutputsAsync(block);
            await block.Completion;
            Assert.IsTrue(results.SequenceEqual(new[] { 13, 17, 19 }));
        }

        [TestMethod]
        public async Task FuncBlock_FuncFaults_Faults()
        {
            var block = new FuncBlock<int>(async send =>
            {
                await send(13);
                throw new NotImplementedException();
            });

            await GetBlockOutputsAsync(block);
            await AssertEx.ThrowsExceptionAsync<NotImplementedException>(block.Completion, allowDerivedTypes: false);
        }

        [TestMethod]
        public async Task FuncBlock_FuncCancels_Completes()
        {
            var cts = new CancellationTokenSource();
            cts.Cancel();
            var block = new FuncBlock<int>(async send =>
            {
                await send(13);
                cts.Token.ThrowIfCancellationRequested();
            });

            var results = await GetBlockOutputsAsync(block);
            await block.Completion;
            Assert.IsTrue(results.SequenceEqual(new[] { 13 }));
        }

        [TestMethod]
        public async Task FuncBlock_Canceled_Cancels()
        {
            var cts = new CancellationTokenSource();
            var block = new FuncBlock<int>(async send =>
            {
                while (true)
                {
                    await send(13);
                    await Task.Delay(100);
                }
            },
            new DataflowBlockOptions
            {
                CancellationToken = cts.Token,
            });

            var resultTask = GetBlockOutputsAsync(block);
            cts.Cancel();
            var results = await resultTask;
            await AssertEx.CompletesCanceledAsync(block.Completion);
        }

        [TestMethod]
        public async Task FuncBlock_Canceled_CancelsFunc()
        {
            var tcs = new TaskCompletionSource();
            var cts = new CancellationTokenSource();
            var block = new FuncBlock<int>(async send =>
            {
                try
                {
                    while (true)
                    {
                        await send(13);
                        await Task.Delay(100);
                    }
                }
                catch (OperationCanceledException)
                {
                    tcs.SetResult();
                }
            },
            new DataflowBlockOptions
            {
                CancellationToken = cts.Token,
            });

            var resultTask = GetBlockOutputsAsync(block);
            cts.Cancel();
            await resultTask;
            await tcs.Task;
        }

        [TestMethod]
        public async Task FuncBlock_Completed_Completes()
        {
            var block = new FuncBlock<int>(async send =>
            {
                while (true)
                {
                    await send(13);
                    await Task.Delay(100);
                }
            });

            var resultTask = GetBlockOutputsAsync(block);
            block.Complete();
            await resultTask;
            await block.Completion;
        }

        [TestMethod]
        public async Task FuncBlock_Completed_CancelsFunc()
        {
            var tcs = new TaskCompletionSource();
            var block = new FuncBlock<int>(async send =>
            {
                try
                {
                    while (true)
                    {
                        await send(13);
                        await Task.Delay(100);
                    }
                }
                catch (OperationCanceledException)
                {
                    tcs.SetResult();
                }
            });

            var resultTask = GetBlockOutputsAsync(block);
            block.Complete();
            await resultTask;
            await block.Completion;
            await tcs.Task;
        }

        [TestMethod]
        public async Task FuncBlock_Faulted_Faults()
        {
            var block = new FuncBlock<int>(async send =>
            {
                while (true)
                {
                    await send(13);
                    await Task.Delay(100);
                }
            });

            ((IDataflowBlock)block).Fault(new NotImplementedException());
            await AssertEx.ThrowsExceptionAsync<NotImplementedException>(block.Completion, allowDerivedTypes: false);
        }

        [TestMethod]
        public async Task FuncBlock_Faulted_CancelsFunc()
        {
            var tcs = new TaskCompletionSource();
            var block = new FuncBlock<int>(async send =>
            {
                try
                {
                    while (true)
                    {
                        await send(13);
                        await Task.Delay(100);
                    }
                }
                catch (OperationCanceledException)
                {
                    tcs.SetResult();
                }
            });

            ((IDataflowBlock)block).Fault(new NotImplementedException());
            await tcs.Task;
        }
    }
}
