﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;
using System.ComponentModel;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class TaskCompletionSourceExtensionsUnitTests_NET45_NE
    {
        [TestMethod]
        public async Task TryCompleteFromCompletedTaskTResult_PropagatesResult()
        {
            var tcs = new TaskCompletionSource<int>();
            tcs.TryCompleteFromCompletedTask(TaskConstants.Int32NegativeOne);
            var result = await tcs.Task;
            Assert.AreEqual(-1, result);
        }

        [TestMethod]
        public async Task TryCompleteFromCompletedTask_WithDifferentTResult_PropagatesResult()
        {
            var tcs = new TaskCompletionSource<object>();
            tcs.TryCompleteFromCompletedTask(TaskConstants.Int32NegativeOne);
            var result = await tcs.Task;
            Assert.AreEqual(-1, result);
        }

        [TestMethod]
        public async Task TryCompleteFromCompletedTaskTResult_PropagatesCancellation()
        {
            var tcs = new TaskCompletionSource<int>();
            tcs.TryCompleteFromCompletedTask(TaskConstants<int>.Canceled);
            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => tcs.Task);
        }

        [TestMethod]
        public async Task TryCompleteFromCompletedTaskTResult_PropagatesException()
        {
            var source = new TaskCompletionSource<int>();
            source.TrySetException(new NotImplementedException());

            var tcs = new TaskCompletionSource<int>();
            tcs.TryCompleteFromCompletedTask(source.Task);
            await AssertEx.ThrowsExceptionAsync<NotImplementedException>(() => tcs.Task);
        }

        [TestMethod]
        public async Task TryCompleteFromCompletedTask_PropagatesResult()
        {
            var tcs = new TaskCompletionSource();
            tcs.TryCompleteFromCompletedTask(TaskConstants.Completed);
            await tcs.Task;
        }

        [TestMethod]
        public async Task TryCompleteFromCompletedTask_PropagatesCancellation()
        {
            var tcs = new TaskCompletionSource();
            tcs.TryCompleteFromCompletedTask(TaskConstants.Canceled);
            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => tcs.Task);
        }

        [TestMethod]
        public async Task TryCompleteFromCompletedTask_PropagatesException()
        {
            var source = new TaskCompletionSource();
            source.TrySetException(new NotImplementedException());

            var tcs = new TaskCompletionSource();
            tcs.TryCompleteFromCompletedTask(source.Task);
            await AssertEx.ThrowsExceptionAsync<NotImplementedException>(() => tcs.Task);
        }

        [TestMethod]
        public async Task TryCompleteFromEventArgsTResult_PropagatesResult()
        {
            var tcs = new TaskCompletionSource<int>();
            tcs.TryCompleteFromEventArgs(new AsyncCompletedEventArgs(null, false, null), () => 13);
            var result = await tcs.Task;
            Assert.AreEqual(13, result);
        }

        [TestMethod]
        public async Task TryCompleteFromEventArgsTResult_PropagatesCancellation()
        {
            var tcs = new TaskCompletionSource<int>();
            tcs.TryCompleteFromEventArgs(new AsyncCompletedEventArgs(null, true, null), null);
            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => tcs.Task);
        }

        [TestMethod]
        public async Task TryCompleteFromEventArgsTResult_PropagatesException()
        {
            var tcs = new TaskCompletionSource<int>();
            tcs.TryCompleteFromEventArgs(new AsyncCompletedEventArgs(new NotImplementedException(), false, null), null);
            await AssertEx.ThrowsExceptionAsync<NotImplementedException>(() => tcs.Task);
        }

        [TestMethod]
        public async Task TryCompleteFromEventArgs_PropagatesResult()
        {
            var tcs = new TaskCompletionSource();
            tcs.TryCompleteFromEventArgs(new AsyncCompletedEventArgs(null, false, null));
            await tcs.Task;
        }

        [TestMethod]
        public async Task TryCompleteFromEventArgs_PropagatesCancellation()
        {
            var tcs = new TaskCompletionSource();
            tcs.TryCompleteFromEventArgs(new AsyncCompletedEventArgs(null, true, null));
            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => tcs.Task);
        }

        [TestMethod]
        public async Task TryCompleteFromEventArgs_PropagatesException()
        {
            var tcs = new TaskCompletionSource();
            tcs.TryCompleteFromEventArgs(new AsyncCompletedEventArgs(new NotImplementedException(), false, null));
            await AssertEx.ThrowsExceptionAsync<NotImplementedException>(() => tcs.Task);
        }

        [TestMethod]
        public async Task TrySetResultWithBackgroundContinuationsTResult_RunsOnAnotherThread()
        {
            var unitTestThreadId = Thread.CurrentThread.ManagedThreadId;
            int continuationThreadId = unitTestThreadId;
            var tcs = new TaskCompletionSource<int>();
            var continuation = tcs.Task.ContinueWith(_ => { continuationThreadId = Thread.CurrentThread.ManagedThreadId; },
                TaskContinuationOptions.ExecuteSynchronously);

            tcs.TrySetResultWithBackgroundContinuations(13);
            await continuation;

            Assert.AreNotEqual(unitTestThreadId, continuationThreadId);
        }

        [TestMethod]
        public async Task TrySetResultWithBackgroundContinuations_RunsOnAnotherThread()
        {
            var unitTestThreadId = Thread.CurrentThread.ManagedThreadId;
            int continuationThreadId = unitTestThreadId;
            var tcs = new TaskCompletionSource();
            var continuation = tcs.Task.ContinueWith(_ => { continuationThreadId = Thread.CurrentThread.ManagedThreadId; },
                TaskContinuationOptions.ExecuteSynchronously);

            tcs.TrySetResultWithBackgroundContinuations();
            await continuation;

            Assert.AreNotEqual(unitTestThreadId, continuationThreadId);
        }

        [TestMethod]
        public async Task TrySetCanceledWithBackgroundContinuationsTResult_RunsOnAnotherThread()
        {
            var unitTestThreadId = Thread.CurrentThread.ManagedThreadId;
            int continuationThreadId = unitTestThreadId;
            var tcs = new TaskCompletionSource<int>();
            var continuation = tcs.Task.ContinueWith(_ => { continuationThreadId = Thread.CurrentThread.ManagedThreadId; },
                TaskContinuationOptions.ExecuteSynchronously);

            tcs.TrySetCanceledWithBackgroundContinuations();
            await continuation;

            Assert.AreNotEqual(unitTestThreadId, continuationThreadId);
        }

        [TestMethod]
        public async Task TrySetCanceledWithBackgroundContinuations_RunsOnAnotherThread()
        {
            var unitTestThreadId = Thread.CurrentThread.ManagedThreadId;
            int continuationThreadId = unitTestThreadId;
            var tcs = new TaskCompletionSource();
            var continuation = tcs.Task.ContinueWith(_ => { continuationThreadId = Thread.CurrentThread.ManagedThreadId; },
                TaskContinuationOptions.ExecuteSynchronously);

            tcs.TrySetCanceledWithBackgroundContinuations();
            await continuation;

            Assert.AreNotEqual(unitTestThreadId, continuationThreadId);
        }

        [TestMethod]
        public async Task TrySetExceptionWithBackgroundContinuationsTResult_RunsOnAnotherThread()
        {
            var unitTestThreadId = Thread.CurrentThread.ManagedThreadId;
            int continuationThreadId = unitTestThreadId;
            var tcs = new TaskCompletionSource<int>();
            var continuation = tcs.Task.ContinueWith(_ => { continuationThreadId = Thread.CurrentThread.ManagedThreadId; },
                TaskContinuationOptions.ExecuteSynchronously);

            tcs.TrySetExceptionWithBackgroundContinuations(new NotImplementedException());
            await continuation;

            Assert.AreNotEqual(unitTestThreadId, continuationThreadId);
        }

        [TestMethod]
        public async Task TrySetExceptionWithBackgroundContinuations_RunsOnAnotherThread()
        {
            var unitTestThreadId = Thread.CurrentThread.ManagedThreadId;
            int continuationThreadId = unitTestThreadId;
            var tcs = new TaskCompletionSource();
            var continuation = tcs.Task.ContinueWith(_ => { continuationThreadId = Thread.CurrentThread.ManagedThreadId; },
                TaskContinuationOptions.ExecuteSynchronously);

            tcs.TrySetExceptionWithBackgroundContinuations(new NotImplementedException());
            await continuation;

            Assert.AreNotEqual(unitTestThreadId, continuationThreadId);
        }

        [TestMethod]
        public async Task TrySetExceptionsWithBackgroundContinuationsTResult_RunsOnAnotherThread()
        {
            var unitTestThreadId = Thread.CurrentThread.ManagedThreadId;
            int continuationThreadId = unitTestThreadId;
            var tcs = new TaskCompletionSource<int>();
            var continuation = tcs.Task.ContinueWith(_ => { continuationThreadId = Thread.CurrentThread.ManagedThreadId; },
                TaskContinuationOptions.ExecuteSynchronously);

            tcs.TrySetExceptionWithBackgroundContinuations(new[] { new NotImplementedException() });
            await continuation;

            Assert.AreNotEqual(unitTestThreadId, continuationThreadId);
        }

        [TestMethod]
        public async Task TrySetExceptionsWithBackgroundContinuations_RunsOnAnotherThread()
        {
            var unitTestThreadId = Thread.CurrentThread.ManagedThreadId;
            int continuationThreadId = unitTestThreadId;
            var tcs = new TaskCompletionSource();
            var continuation = tcs.Task.ContinueWith(_ => { continuationThreadId = Thread.CurrentThread.ManagedThreadId; },
                TaskContinuationOptions.ExecuteSynchronously);

            tcs.TrySetExceptionWithBackgroundContinuations(new[] { new NotImplementedException() });
            await continuation;

            Assert.AreNotEqual(unitTestThreadId, continuationThreadId);
        }
    }
}
