﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using Nito.AsyncEx.Synchronous;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;
using Etw.AsyncAutoResetEvent;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncAutoResetEventUnitTests_NET45
    {
        [TestMethod]
        public async Task WaitAsync_Unset_IsNotCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent();

            var task = are.WaitAsync();

            await AssertEx.NeverCompletesAsync(task);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectWaitBegin(task.Id, IsSynchronous: false)
                .ExpectDone();
        }

        [TestMethod]
        public void WaitAsync_AfterSet_CompletesSynchronously()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent();
            
            are.Set();
            var task = are.WaitAsync();
            
            Assert.IsTrue(task.IsCompleted);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectSet()
                .ExpectWaitBegin(task.Id, IsSynchronous: true)
                .ExpectWaitEnd(task.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public void WaitAsync_Set_CompletesSynchronously()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent(true);

            var task = are.WaitAsync();
            
            Assert.IsTrue(task.IsCompleted);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectSet()
                .ExpectWaitBegin(task.Id, IsSynchronous: true)
                .ExpectWaitEnd(task.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task MultipleWaitAsync_AfterSet_OnlyOneIsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent();

            are.Set();
            var task1 = are.WaitAsync();
            var task2 = are.WaitAsync();
            
            Assert.IsTrue(task1.IsCompleted);
            await AssertEx.NeverCompletesAsync(task2);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectSet()
                .ExpectWaitBegin(task1.Id, IsSynchronous: true)
                .ExpectWaitEnd(task1.Id, IsCanceled: false)
                .ExpectWaitBegin(task2.Id, IsSynchronous: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task MultipleWaitAsync_Set_OnlyOneIsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent(true);

            var task1 = are.WaitAsync();
            var task2 = are.WaitAsync();
            
            Assert.IsTrue(task1.IsCompleted);
            await AssertEx.NeverCompletesAsync(task2);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectSet()
                .ExpectWaitBegin(task1.Id, IsSynchronous: true)
                .ExpectWaitEnd(task1.Id, IsCanceled: false)
                .ExpectWaitBegin(task2.Id, IsSynchronous: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task MultipleWaitAsync_AfterMultipleSet_OnlyOneIsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent();

            are.Set();
            are.Set();
            var task1 = are.WaitAsync();
            var task2 = are.WaitAsync();
            
            Assert.IsTrue(task1.IsCompleted);
            await AssertEx.NeverCompletesAsync(task2);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectSet()
                .ExpectSet()
                .ExpectWaitBegin(task1.Id, IsSynchronous: true)
                .ExpectWaitEnd(task1.Id, IsCanceled: false)
                .ExpectWaitBegin(task2.Id, IsSynchronous: false)
                .ExpectDone();
        }

        [TestMethod]
        public void WaitAsync_PreCancelled_Set_SynchronouslyCompletesWait()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent(true);
            var token = new CancellationToken(true);
            
            var task = are.WaitAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsFalse(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectSet()
                .ExpectWaitBegin(task.Id, IsSynchronous: true)
                .ExpectWaitEnd(task.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task WaitAsync_Cancelled_DoesNotAutoReset()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent();
            var cts = new CancellationTokenSource();

            cts.Cancel();
            var task1 = are.WaitAsync(cts.Token);
            task1.WaitWithoutException();
            are.Set();
            var task2 = are.WaitAsync();
            
            await task2;
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectWaitBegin(task1.Id, IsSynchronous: true)
                .ExpectWaitEnd(task1.Id, IsCanceled: true)
                .ExpectSet()
                .ExpectWaitBegin(task2.Id, IsSynchronous: true)
                .ExpectWaitEnd(task2.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public void WaitAsync_PreCancelled_Unset_SynchronouslyCancels()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent(false);
            var token = new CancellationToken(true);

            var task = are.WaitAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsTrue(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectWaitBegin(task.Id, IsSynchronous: true)
                .ExpectWaitEnd(task.Id, IsCanceled: true)
                .ExpectDone();
        }

        [TestMethod]
        public void WaitAsyncFromCustomSynchronizationContext_PreCancelled_Unset_SynchronouslyCancels()
        {
            AsyncContext.Run(() =>
            {
                var etw = DebugAsyncExEtw.Start();
                var are = new AsyncAutoResetEvent(false);
                var token = new CancellationToken(true);

                var task = are.WaitAsync(token);

                Assert.IsTrue(task.IsCompleted);
                Assert.IsTrue(task.IsCanceled);
                Assert.IsFalse(task.IsFaulted);
                etw.StopForAsyncAutoResetEvent(are.Id)
                    .ExpectWaitBegin(task.Id, IsSynchronous: true)
                    .ExpectWaitEnd(task.Id, IsCanceled: true)
                    .ExpectDone();
            });
        }

        [TestMethod]
        public async Task WaitAsync_Cancelled_ThrowsException()
        {
            var etw = DebugAsyncExEtw.Start();
            var are = new AsyncAutoResetEvent();
            var cts = new CancellationTokenSource();
            cts.Cancel();
            var task = are.WaitAsync(cts.Token);
            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(task);
            etw.StopForAsyncAutoResetEvent(are.Id)
                .ExpectWaitBegin(task.Id, IsSynchronous: true)
                .ExpectWaitEnd(task.Id, IsCanceled: true)
                .ExpectDone();
        }

        [TestMethod]
        public void Id_IsNotZero()
        {
            var are = new AsyncAutoResetEvent();
            Assert.AreNotEqual(0, are.Id);
        }
    }
}

namespace Etw.AsyncAutoResetEvent
{
    [ExcludeFromCodeCoverage]
    internal static partial class EtwExpect
    {
        private static readonly ExpectedEtwEvent WaitBegin = new ExpectedEtwEvent("Nito.AsyncEx", 19, 0);
        private static readonly ExpectedEtwEvent WaitEnd = new ExpectedEtwEvent("Nito.AsyncEx", 20, 0);
        private static readonly ExpectedEtwEvent Set = new ExpectedEtwEvent("Nito.AsyncEx", 21, 0);

        public static DebugEtw.EventCollection ExpectWaitBegin(this DebugEtw.EventCollection events, int WaitTaskID, bool IsSynchronous)
        {
            return events.Expect(WaitBegin, new { events.Bag.AsyncAutoResetEventID, WaitTaskID, IsSynchronous });
        }

        public static DebugEtw.EventCollection ExpectWaitEnd(this DebugEtw.EventCollection events, int WaitTaskID, bool IsCanceled)
        {
            return events.Expect(WaitEnd, new { events.Bag.AsyncAutoResetEventID, WaitTaskID, IsCanceled });
        }

        public static DebugEtw.EventCollection ExpectSet(this DebugEtw.EventCollection events)
        {
            return events.Expect(Set, new { events.Bag.AsyncAutoResetEventID });
        }

        public static DebugEtw.EventCollection StopForAsyncAutoResetEvent(this DebugEtw etw, int AsyncAutoResetEventID)
        {
            var ret = etw.Stop(e => e.DataMatches(new { AsyncAutoResetEventID }));
            ret.Bag.AsyncAutoResetEventID = AsyncAutoResetEventID;
            return ret;
        }
    }
}
