﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;
using Etw.AsyncConditionVariable;
using Nito.AsyncEx.Internal;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncConditionVariableUnitTests_NET45
    {
        [TestMethod]
        public async Task WaitAsync_WithoutNotify_IsNotCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var cv = new AsyncConditionVariable(mutex);

            await mutex.LockAsync();
            var task = cv.WaitAsync();

            await AssertEx.NeverCompletesAsync(task);
            etw.StopForAsyncConditionVariable(cv.Id)
                .ExpectWaitBegin(task.Id)
                .ExpectDone();
        }

        [TestMethod]
        public async Task WaitAsync_Notified_IsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var cv = new AsyncConditionVariable(mutex);
            await mutex.LockAsync();
            var task = cv.WaitAsync();

            await Task.Run(async () =>
            {
                using (await mutex.LockAsync())
                {
                    cv.Notify();
                }
            });
            await task;

            etw.StopForAsyncConditionVariable(cv.Id)
                .ExpectWaitBegin(task.Id)
                .ExpectNotifyOne()
                .ExpectWaitEnd(task.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task WaitAsync_AfterNotify_IsNotCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var cv = new AsyncConditionVariable(mutex);
            await Task.Run(async () =>
            {
                using (await mutex.LockAsync())
                {
                    cv.Notify();
                }
            });

            await mutex.LockAsync();
            var task = cv.WaitAsync();

            await AssertEx.NeverCompletesAsync(task);
            etw.StopForAsyncConditionVariable(cv.Id)
                .ExpectNotifyOne()
                .ExpectWaitBegin(task.Id)
                .ExpectDone();
        }

        [TestMethod]
        public async Task MultipleWaits_NotifyAll_AllAreCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var cv = new AsyncConditionVariable(mutex);
            var key1 = await mutex.LockAsync();
            var task1 = cv.WaitAsync();
            var __ = task1.ContinueWith(_ => key1.Dispose());
            var key2 = await mutex.LockAsync();
            var task2 = cv.WaitAsync();
            var ___ = task2.ContinueWith(_ => key2.Dispose());

            await Task.Run(async () =>
            {
                using (await mutex.LockAsync())
                {
                    cv.NotifyAll();
                }
            });
            
            await task1;
            await task2;
            etw.StopForAsyncConditionVariable(cv.Id)
                .ExpectWaitBegin(task1.Id)
                .ExpectWaitBegin(task2.Id)
                .ExpectNotifyAll()
                .ExpectWaitEnd(task1.Id, IsCanceled: false)
                .ExpectWaitEnd(task2.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task MultipleWaits_Notify_OneIsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var cv = new AsyncConditionVariable(mutex);
            var key = await mutex.LockAsync();
            var task1 = cv.WaitAsync();
            var __ = task1.ContinueWith(_ => key.Dispose());
            await mutex.LockAsync();
            var task2 = cv.WaitAsync();

            await Task.Run(async () =>
            {
                using (await mutex.LockAsync())
                {
                    cv.Notify();
                }
            });

            await task1;
            await AssertEx.NeverCompletesAsync(task2);
            etw.StopForAsyncConditionVariable(cv.Id)
                .ExpectWaitBegin(task1.Id)
                .ExpectWaitBegin(task2.Id)
                .ExpectNotifyOne()
                .ExpectWaitEnd(task1.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public void Id_IsNotZero()
        {
            var mutex = new AsyncLock();
            var cv = new AsyncConditionVariable(mutex);
            Assert.AreNotEqual(0, cv.Id);
        }
    }
}

namespace Etw.AsyncConditionVariable
{
    [ExcludeFromCodeCoverage]
    internal static partial class EtwExpect
    {
        private static readonly ExpectedEtwEvent NotifyOne = new ExpectedEtwEvent("Nito.AsyncEx", 14, 0);
        private static readonly ExpectedEtwEvent NotifyAll = new ExpectedEtwEvent("Nito.AsyncEx", 15, 0);
        private static readonly ExpectedEtwEvent WaitBegin = new ExpectedEtwEvent("Nito.AsyncEx", 16, 0);
        private static readonly ExpectedEtwEvent WaitEnd = new ExpectedEtwEvent("Nito.AsyncEx", 17, 0);

        public static DebugEtw.EventCollection ExpectNotifyOne(this DebugEtw.EventCollection events)
        {
            return events.Expect(NotifyOne, new { events.Bag.AsyncConditionVariableID });
        }

        public static DebugEtw.EventCollection ExpectNotifyAll(this DebugEtw.EventCollection events)
        {
            return events.Expect(NotifyAll, new { events.Bag.AsyncConditionVariableID });
        }

        public static DebugEtw.EventCollection ExpectWaitBegin(this DebugEtw.EventCollection events, int WaitTaskID)
        {
            return events.Expect(WaitBegin, new { events.Bag.AsyncConditionVariableID, WaitTaskID });
        }

        public static DebugEtw.EventCollection ExpectWaitEnd(this DebugEtw.EventCollection events, int WaitTaskID, bool IsCanceled)
        {
            return events.Expect(WaitEnd, new { events.Bag.AsyncConditionVariableID, WaitTaskID, IsCanceled });
        }

        public static DebugEtw.EventCollection StopForAsyncConditionVariable(this DebugEtw etw, int AsyncConditionVariableID)
        {
            var ret = etw.Stop(e => e.DataMatches(new { AsyncConditionVariableID }));
            ret.Bag.AsyncConditionVariableID = AsyncConditionVariableID;
            return ret;
        }
    }
}
