﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;
using Etw.AsyncLock;
using Nito.AsyncEx.Internal;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncLockUnitTests_NET45
    {
        [TestMethod]
        public void AsyncLock_Unlocked_SynchronouslyPermitsLock()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();

            var lockTask = mutex.LockAsync();

            Assert.IsTrue(lockTask.IsCompleted);
            Assert.IsFalse(lockTask.IsFaulted);
            Assert.IsFalse(lockTask.IsCanceled);
            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: lockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: lockTask.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task AsyncLock_Locked_PreventsLockUntilUnlocked()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var task1HasLock = new TaskCompletionSource();
            var task1Continue = new TaskCompletionSource();

            Task<IDisposable> task1LockTask = null;
            var task1 = Task.Run(async () =>
            {
                task1LockTask = mutex.LockAsync();
                using (await task1LockTask)
                {
                    task1HasLock.SetResult();
                    await task1Continue.Task;
                }
            });
            await task1HasLock.Task;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task1LockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: task1LockTask.Id, IsCanceled: false)
                .ExpectDone();

            Task<IDisposable> task2LockTask = null;
            var task2Start = Task.Factory.StartNew(async () =>
            {
                task2LockTask = mutex.LockAsync();
                await task2LockTask;
            });
            var task2 = await task2Start;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task2LockTask.Id, IsSynchronous: false)
                .ExpectDone();

            Assert.IsFalse(task2.IsCompleted);
            task1Continue.SetResult();
            await task2;
            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectUnlocked()
                .ExpectLockEnd(LockTaskID: task2LockTask.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task AsyncLock_DoubleDispose_OnlyPermitsOneTask()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var task1HasLock = new TaskCompletionSource();
            var task1Continue = new TaskCompletionSource();

            Task<IDisposable> lockTask = null;
            await Task.Run(async () =>
            {
                lockTask = mutex.LockAsync();
                var key = await lockTask;
                key.Dispose();
                key.Dispose();
            });
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: lockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: lockTask.Id, IsCanceled: false)
                .ExpectUnlocked()
                .ExpectDone();

            Task<IDisposable> task1LockTask = null;
            var task1 = Task.Run(async () =>
            {
                task1LockTask = mutex.LockAsync();
                using (await task1LockTask)
                {
                    task1HasLock.SetResult();
                    await task1Continue.Task;
                }
            });
            await task1HasLock.Task;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task1LockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: task1LockTask.Id, IsCanceled: false)
                .ExpectDone();

            Task<IDisposable> task2LockTask = null;
            var task2Start = Task.Factory.StartNew(async () =>
            {
                task2LockTask = mutex.LockAsync();
                await task2LockTask;
            });
            var task2 = await task2Start;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task2LockTask.Id, IsSynchronous: false)
                .ExpectDone();

            Assert.IsFalse(task2.IsCompleted);
            task1Continue.SetResult();
            await task2;

            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectUnlocked()
                .ExpectLockEnd(LockTaskID: task2LockTask.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task AsyncLock_Locked_OnlyPermitsOneLockerAtATime()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var task1HasLock = new TaskCompletionSource();
            var task1Continue = new TaskCompletionSource();
            var task2HasLock = new TaskCompletionSource();
            var task2Continue = new TaskCompletionSource();
            Task<IDisposable> task1LockTask = null, task2LockTask = null, task3LockTask = null;

            var task1 = Task.Run(async () =>
            {
                task1LockTask = mutex.LockAsync();
                using (await task1LockTask)
                {
                    task1HasLock.SetResult();
                    await task1Continue.Task;
                }
            });
            await task1HasLock.Task;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task1LockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: task1LockTask.Id, IsCanceled: false)
                .ExpectDone();

            var task2Start = Task.Factory.StartNew(async () =>
            {
                task2LockTask = mutex.LockAsync();
                using (await task2LockTask)
                {
                    task2HasLock.SetResult();
                    await task2Continue.Task;
                }
            });
            var task2 = await task2Start;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task2LockTask.Id, IsSynchronous: false)
                .ExpectDone();

            var task3Start = Task.Factory.StartNew(async () =>
            {
                task3LockTask = mutex.LockAsync();
                await task3LockTask;
            });
            var task3 = await task3Start;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task3LockTask.Id, IsSynchronous: false)
                .ExpectDone();

            task1Continue.SetResult();
            await task2HasLock.Task;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectUnlocked()
                .ExpectLockEnd(LockTaskID: task2LockTask.Id, IsCanceled: false)
                .ExpectDone();

            Assert.IsFalse(task3.IsCompleted);
            task2Continue.SetResult();
            await task2;
            await task3;
            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectUnlocked()
                .ExpectLockEnd(LockTaskID: task3LockTask.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public void AsyncLock_PreCancelled_Unlocked_SynchronouslyTakesLock()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var token = new CancellationToken(true);

            var task = mutex.LockAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsFalse(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: task.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public void AsyncLock_PreCancelled_Locked_SynchronouslyCancels()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var lockTask = mutex.LockAsync();
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: lockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: lockTask.Id, IsCanceled: false)
                .ExpectDone();
            var token = new CancellationToken(true);

            var task = mutex.LockAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsTrue(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: task.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: task.Id, IsCanceled: true)
                .ExpectDone();
        }

        [TestMethod]
        public async Task AsyncLock_CancelledLock_LeavesLockUnlocked()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var cts = new CancellationTokenSource();
            var taskReady = new TaskCompletionSource();
            Task<IDisposable> initialLockTask = null, lockTask = null;

            initialLockTask = mutex.LockAsync();
            var unlock = await initialLockTask;
            var task = Task.Run(async () =>
            {
                lockTask = mutex.LockAsync(cts.Token);
                taskReady.SetResult();
                await lockTask;
            });
            await taskReady.Task;
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: initialLockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: initialLockTask.Id, IsCanceled: false)
                .ExpectLockBegin(LockTaskID: lockTask.Id, IsSynchronous: false)
                .ExpectDone();
            cts.Cancel();
            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(task);
            Assert.IsTrue(task.IsCanceled);
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockEnd(LockTaskID: lockTask.Id, IsCanceled: true)
                .ExpectDone();
            unlock.Dispose();
            etw.SnapshotForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectUnlocked()
                .ExpectDone();

            var finalLockTask = mutex.LockAsync();
            await finalLockTask;
            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: finalLockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: finalLockTask.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task AsyncLock_CanceledLock_ThrowsException()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var cts = new CancellationTokenSource();

            var initialLockTask = mutex.LockAsync();
            await initialLockTask;
            var canceledLockTask = mutex.LockAsync(cts.Token);
            cts.Cancel();

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(canceledLockTask);
            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: initialLockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: initialLockTask.Id, IsCanceled: false)
                .ExpectLockBegin(LockTaskID: canceledLockTask.Id, IsSynchronous: false)
                .ExpectLockEnd(LockTaskID: canceledLockTask.Id, IsCanceled: true)
                .ExpectDone();
        }

        [TestMethod]
        public async Task AsyncLock_CanceledTooLate_StillTakesLock()
        {
            var etw = DebugAsyncExEtw.Start();
            var mutex = new AsyncLock();
            var cts = new CancellationTokenSource();

            Task<IDisposable> initialLockTask, cancelableLockTask;
            initialLockTask = mutex.LockAsync();
            using (await initialLockTask)
            {
                cancelableLockTask = mutex.LockAsync(cts.Token);
            }

            var key = await cancelableLockTask;
            cts.Cancel();

            var nextLocker = mutex.LockAsync();
            Assert.IsFalse(nextLocker.IsCompleted);

            key.Dispose();
            await nextLocker;
            etw.StopForAsyncLock(AsyncLockID: mutex.Id)
                .ExpectLockBegin(LockTaskID: initialLockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(LockTaskID: initialLockTask.Id, IsCanceled: false)
                .ExpectLockBegin(LockTaskID: cancelableLockTask.Id, IsSynchronous: false)
                .ExpectUnlocked()
                .ExpectLockEnd(LockTaskID: cancelableLockTask.Id, IsCanceled: false)
                .ExpectLockBegin(LockTaskID: nextLocker.Id, IsSynchronous: false)
                .ExpectUnlocked()
                .ExpectLockEnd(LockTaskID: nextLocker.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public void Id_IsNotZero()
        {
            var mutex = new AsyncLock();
            Assert.AreNotEqual(0, mutex.Id);
        }
    }
}

namespace Etw.AsyncLock
{
    [ExcludeFromCodeCoverage]
    internal static partial class EtwExpect
    {
        private static readonly ExpectedEtwEvent AsyncLockBegin = new ExpectedEtwEvent("Nito.AsyncEx", 7, 0);
        private static readonly ExpectedEtwEvent AsyncLockEnd = new ExpectedEtwEvent("Nito.AsyncEx", 8, 0);
        private static readonly ExpectedEtwEvent AsyncLockUnlocked = new ExpectedEtwEvent("Nito.AsyncEx", 9, 0);

        public static DebugEtw.EventCollection ExpectLockBegin(this DebugEtw.EventCollection events, int? LockTaskID, bool IsSynchronous)
        {
            return events.Expect(AsyncLockBegin, new { events.Bag.AsyncLockID, LockTaskID, IsSynchronous });
        }

        public static DebugEtw.EventCollection ExpectLockEnd(this DebugEtw.EventCollection events, int? LockTaskID, bool IsCanceled)
        {
            return events.Expect(AsyncLockEnd, new { events.Bag.AsyncLockID, LockTaskID, IsCanceled });
        }

        public static DebugEtw.EventCollection ExpectUnlocked(this DebugEtw.EventCollection events)
        {
            return events.Expect(AsyncLockUnlocked, new { events.Bag.AsyncLockID });
        }

        public static DebugEtw.EventCollection StopForAsyncLock(this DebugEtw etw, int AsyncLockID)
        {
            var ret = etw.Stop(e => e.DataMatches(new { AsyncLockID }));
            ret.Bag.AsyncLockID = AsyncLockID;
            return ret;
        }

        public static DebugEtw.EventCollection SnapshotForAsyncLock(this DebugEtw etw, int AsyncLockID)
        {
            var ret = etw.Snapshot(e => e.DataMatches(new { AsyncLockID }));
            ret.Bag.AsyncLockID = AsyncLockID;
            return ret;
        }
    }
}