﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;
using Etw.AsyncManualResetEvent;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncManualResetEventUnitTests_NET45
    {
        [TestMethod]
        public async Task WaitAsync_Unset_IsNotCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mre = new AsyncManualResetEvent();

            var task = mre.WaitAsync();

            await AssertEx.NeverCompletesAsync(task);
            etw.StopForAsyncManualResetEvent(AsyncManualResetEventID: mre.Id)
                .ExpectReset(task.Id)
                .ExpectWait(WaitTaskID: task.Id)
                .ExpectDone();
        }

        [TestMethod]
        public void WaitAsync_AfterSet_IsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mre = new AsyncManualResetEvent();

            mre.Set();
            var task = mre.WaitAsync();
            
            Assert.IsTrue(task.IsCompleted);
            etw.StopForAsyncManualResetEvent(AsyncManualResetEventID: mre.Id)
                .ExpectReset(task.Id)
                .ExpectSet(WaitTaskID: task.Id)
                .ExpectWait(WaitTaskID: task.Id)
                .ExpectDone();
        }

        [TestMethod]
        public void WaitAsync_Set_IsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mre = new AsyncManualResetEvent(true);

            var task = mre.WaitAsync();
            
            Assert.IsTrue(task.IsCompleted);
            etw.StopForAsyncManualResetEvent(AsyncManualResetEventID: mre.Id)
                .ExpectSet(WaitTaskID: task.Id)
                .ExpectWait(WaitTaskID: task.Id)
                .ExpectDone();
        }

        [TestMethod]
        public void MultipleWaitAsync_AfterSet_IsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mre = new AsyncManualResetEvent();

            mre.Set();
            var task1 = mre.WaitAsync();
            var task2 = mre.WaitAsync();
            
            Assert.IsTrue(task1.IsCompleted);
            Assert.IsTrue(task2.IsCompleted);
            etw.StopForAsyncManualResetEvent(AsyncManualResetEventID: mre.Id)
                .ExpectReset(task1.Id)
                .ExpectSet(WaitTaskID: task1.Id)
                .ExpectWait(WaitTaskID: task1.Id)
                .ExpectWait(WaitTaskID: task2.Id)
                .ExpectDone();
        }

        [TestMethod]
        public void MultipleWaitAsync_Set_IsCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mre = new AsyncManualResetEvent(true);

            var task1 = mre.WaitAsync();
            var task2 = mre.WaitAsync();
            
            Assert.IsTrue(task1.IsCompleted);
            Assert.IsTrue(task2.IsCompleted);
            etw.StopForAsyncManualResetEvent(AsyncManualResetEventID: mre.Id)
                .ExpectSet(WaitTaskID: task1.Id)
                .ExpectWait(WaitTaskID: task1.Id)
                .ExpectWait(WaitTaskID: task2.Id)
                .ExpectDone();
        }

        [TestMethod]
        public async Task WaitAsync_AfterReset_IsNotCompleted()
        {
            var etw = DebugAsyncExEtw.Start();
            var mre = new AsyncManualResetEvent();

            mre.Set();
            mre.Reset();
            var task = mre.WaitAsync();

            await AssertEx.NeverCompletesAsync(task);
            etw.StopForAsyncManualResetEvent(AsyncManualResetEventID: mre.Id)
                .ExpectReset()
                .ExpectSet()
                .ExpectReset(task.Id)
                .ExpectWait(WaitTaskID: task.Id)
                .ExpectDone();
        }

        [TestMethod]
        public void Id_IsNotZero()
        {
            var mre = new AsyncManualResetEvent();
            Assert.AreNotEqual(0, mre.Id);
        }
    }
}

namespace Etw.AsyncManualResetEvent
{
    [ExcludeFromCodeCoverage]
    internal static partial class EtwExpect
    {
        private static readonly ExpectedEtwEvent Set = new ExpectedEtwEvent("Nito.AsyncEx", 10, 0);
        private static readonly ExpectedEtwEvent Reset = new ExpectedEtwEvent("Nito.AsyncEx", 11, 0);
        private static readonly ExpectedEtwEvent Wait = new ExpectedEtwEvent("Nito.AsyncEx", 12, 0);

        public static DebugEtw.EventCollection ExpectSet(this DebugEtw.EventCollection events, int? WaitTaskID = null)
        {
            return events.Expect(Set, new { events.Bag.AsyncManualResetEventID, WaitTaskID });
        }

        public static DebugEtw.EventCollection ExpectReset(this DebugEtw.EventCollection events, int? WaitTaskID = null)
        {
            return events.Expect(Reset, new { events.Bag.AsyncManualResetEventID, WaitTaskID });
        }

        public static DebugEtw.EventCollection ExpectWait(this DebugEtw.EventCollection events, int WaitTaskID)
        {
            return events.Expect(Wait, new { events.Bag.AsyncManualResetEventID, WaitTaskID });
        }

        public static DebugEtw.EventCollection StopForAsyncManualResetEvent(this DebugEtw etw, int AsyncManualResetEventID)
        {
            var ret = etw.Stop(e => e.DataMatches(new { AsyncManualResetEventID }));
            ret.Bag.AsyncManualResetEventID = AsyncManualResetEventID;
            return ret;
        }
    }
}
