﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;
using Etw.AsyncMonitor;
using Nito.AsyncEx.Internal;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncMonitorUnitTests_NET45
    {
        [TestMethod]
        public async Task Unlocked_PermitsLock()
        {
            var etw = DebugAsyncExEtw.Start();
            var monitor = new AsyncMonitor();
            
            var task = monitor.EnterAsync();
            await task;

            etw.StopForAsyncMonitor(monitor.Id)
                .ExpectLockBegin(task.Id, IsSynchronous: true)
                .ExpectLockEnd(task.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task Locked_PreventsLockUntilUnlocked()
        {
            var etw = DebugAsyncExEtw.Start();
            var monitor = new AsyncMonitor();
            var task1HasLock = new TaskCompletionSource();
            var task1Continue = new TaskCompletionSource();
            Task<IDisposable> initialLockTask = null;

            var task1 = Task.Run(async () =>
            {
                initialLockTask = monitor.EnterAsync();
                using (await initialLockTask)
                {
                    task1HasLock.SetResult();
                    await task1Continue.Task;
                }
            });
            await task1HasLock.Task;

            var lockTask = monitor.EnterAsync();
            Assert.IsFalse(lockTask.IsCompleted);
            task1Continue.SetResult();
            await lockTask;
            etw.StopForAsyncMonitor(monitor.Id)
                .ExpectLockBegin(initialLockTask.Id, IsSynchronous: true)
                .ExpectLockEnd(initialLockTask.Id, IsCanceled: false)
                .ExpectLockBegin(lockTask.Id, IsSynchronous: false)
                .ExpectUnlocked()
                .ExpectLockEnd(lockTask.Id, IsCanceled: false)
                .ExpectDone();
        }

        [TestMethod]
        public async Task Pulse_ReleasesOneWaiter()
        {
            var etw = DebugAsyncExEtw.Start();
            var monitor = new AsyncMonitor();
            int completed = 0;
            var task1Ready = new TaskCompletionSource();
            var task2Ready = new TaskCompletionSource();
            Task<IDisposable> lockTask1 = null;
            Task waitTask1 = null;
            var task1 = Task.Run(async () =>
            {
                lockTask1 = monitor.EnterAsync();
                using (await lockTask1)
                {
                    waitTask1 = monitor.WaitAsync();
                    task1Ready.SetResult();
                    await waitTask1;
                    Interlocked.Increment(ref completed);
                }
            });
            await task1Ready.Task;
            Task<IDisposable> lockTask2 = null;
            Task waitTask2 = null;
            var task2 = Task.Run(async () =>
            {
                lockTask2 = monitor.EnterAsync();
                using (await lockTask2)
                {
                    waitTask2 = monitor.WaitAsync();
                    task2Ready.SetResult();
                    await waitTask2;
                    Interlocked.Increment(ref completed);
                }
            });
            await task2Ready.Task;

            Task<IDisposable> lockTask3 = monitor.EnterAsync();
            using (await lockTask3)
            {
                monitor.Pulse();
            }
            await Task.WhenAny(task1, task2);
            var result = Interlocked.CompareExchange(ref completed, 0, 0);

            Assert.AreEqual(1, result);
            etw.StopForAsyncMonitor(monitor.Id)
                .ExpectLockBegin(lockTask1.Id, IsSynchronous: true)
                .ExpectLockEnd(lockTask1.Id, IsCanceled: false)
                .ExpectWaitBegin(waitTask1.Id)
                .ExpectUnlocked()
                .ExpectLockBegin(lockTask2.Id, IsSynchronous: true)
                .ExpectLockEnd(lockTask2.Id, IsCanceled: false)
                .ExpectWaitBegin(waitTask2.Id)
                .ExpectUnlocked()
                .ExpectLockBegin(lockTask3.Id, IsSynchronous: true)
                .ExpectLockEnd(lockTask3.Id, IsCanceled: false)
                .ExpectNotifyOne()
                .ExpectLockBegin(null, IsSynchronous: false)
                .ExpectUnlocked()
                .ExpectLockEnd(null, IsCanceled: false)
                .ExpectWaitEnd(waitTask1.Id, IsCanceled: false)
                .ExpectUnlocked()
                .ExpectDone();
        }

        [TestMethod]
        public async Task PulseAll_ReleasesAllWaiters()
        {
            var etw = DebugAsyncExEtw.Start();
            var monitor = new AsyncMonitor();
            int completed = 0;
            var task1Ready = new TaskCompletionSource();
            var task2Ready = new TaskCompletionSource();
            Task<IDisposable> lockTask1 = null;
            Task waitTask1 = null;
            var task1 = Task.Run(async () =>
            {
                lockTask1 = monitor.EnterAsync();
                using (await lockTask1)
                {
                    waitTask1 = monitor.WaitAsync();
                    task1Ready.SetResult();
                    await waitTask1;
                    Interlocked.Increment(ref completed);
                }
            });
            await task1Ready.Task;
            Task<IDisposable> lockTask2 = null;
            Task waitTask2 = null;
            var task2 = Task.Run(async () =>
            {
                lockTask2 = monitor.EnterAsync();
                using (await lockTask2)
                {
                    waitTask2 = monitor.WaitAsync();
                    task2Ready.SetResult();
                    await waitTask2;
                    Interlocked.Increment(ref completed);
                }
            });
            await task2Ready.Task;

            var lockTask3 = monitor.EnterAsync();
            using (await lockTask3)
            {
                monitor.PulseAll();
            }
            await Task.WhenAll(task1, task2);
            var result = Interlocked.CompareExchange(ref completed, 0, 0);
            
            Assert.AreEqual(2, result);
            etw.StopForAsyncMonitor(monitor.Id)
                .ExpectLockBegin(lockTask1.Id, IsSynchronous: true)
                .ExpectLockEnd(lockTask1.Id, IsCanceled: false)
                .ExpectWaitBegin(waitTask1.Id)
                .ExpectUnlocked()
                .ExpectLockBegin(lockTask2.Id, IsSynchronous: true)
                .ExpectLockEnd(lockTask2.Id, IsCanceled: false)
                .ExpectWaitBegin(waitTask2.Id)
                .ExpectUnlocked()
                .ExpectLockBegin(lockTask3.Id, IsSynchronous: true)
                .ExpectLockEnd(lockTask3.Id, IsCanceled: false)
                .ExpectNotifyAll()
                .ExpectLockBegin(null, IsSynchronous: false)
                .ExpectLockBegin(null, IsSynchronous: false)
                .ExpectUnlocked()
                .ExpectLockEnd(null, IsCanceled: false)
                .ExpectWaitEnd(waitTask1.Id, IsCanceled: false)
                .ExpectUnlocked()
                .ExpectLockEnd(null, IsCanceled: false)
                .ExpectWaitEnd(waitTask2.Id, IsCanceled: false)
                .ExpectUnlocked()
                .ExpectDone();
        }

        [TestMethod]
        public void Id_IsNotZero()
        {
            var monitor = new AsyncMonitor();
            Assert.AreNotEqual(0, monitor.Id);
        }
    }
}

namespace Etw.AsyncMonitor
{
    [ExcludeFromCodeCoverage]
    internal static partial class EtwExpect
    {
        private static readonly ExpectedEtwEvent Created = new ExpectedEtwEvent("Nito.AsyncEx", 18, 0);

        public static DebugEtw.EventCollection ExpectNotifyOne(this DebugEtw.EventCollection events)
        {
            return Etw.AsyncConditionVariable.EtwExpect.ExpectNotifyOne(events);
        }

        public static DebugEtw.EventCollection ExpectNotifyAll(this DebugEtw.EventCollection events)
        {
            return Etw.AsyncConditionVariable.EtwExpect.ExpectNotifyAll(events);
        }

        public static DebugEtw.EventCollection ExpectWaitBegin(this DebugEtw.EventCollection events, int WaitTaskID)
        {
            return Etw.AsyncConditionVariable.EtwExpect.ExpectWaitBegin(events, WaitTaskID);
        }

        public static DebugEtw.EventCollection ExpectWaitEnd(this DebugEtw.EventCollection events, int WaitTaskID, bool IsCanceled)
        {
            return Etw.AsyncConditionVariable.EtwExpect.ExpectWaitEnd(events, WaitTaskID, IsCanceled);
        }

        public static DebugEtw.EventCollection ExpectLockBegin(this DebugEtw.EventCollection events, int? LockTaskID, bool IsSynchronous)
        {
            return Etw.AsyncLock.EtwExpect.ExpectLockBegin(events, LockTaskID, IsSynchronous);
        }

        public static DebugEtw.EventCollection ExpectLockEnd(this DebugEtw.EventCollection events, int? LockTaskID, bool IsCanceled)
        {
            return Etw.AsyncLock.EtwExpect.ExpectLockEnd(events, LockTaskID, IsCanceled);
        }

        public static DebugEtw.EventCollection ExpectUnlocked(this DebugEtw.EventCollection events)
        {
            return Etw.AsyncLock.EtwExpect.ExpectUnlocked(events);
        }

        private static bool MatchesLock(DebugEtw.Event a, DebugEtw.Event b)
        {
            return (a.DataNames.Contains("AsyncLockID") && b.DataNames.Contains("AsyncLockID") && a.Data.AsyncLockID == b.Data.AsyncLockID);
        }

        private static bool MatchesCV(DebugEtw.Event a, DebugEtw.Event b)
        {
            return (a.DataNames.Contains("AsyncConditionVariableID") && b.DataNames.Contains("AsyncConditionVariableID") && a.Data.AsyncConditionVariableID == b.Data.AsyncConditionVariableID);
        }

        public static DebugEtw.EventCollection StopForAsyncMonitor(this DebugEtw etw, int AsyncMonitorID)
        {
            var ret = etw.Stop(e => e.DataMatches(new { AsyncLockID = AsyncMonitorID }), (e1, e2) => MatchesLock(e1, e2) || MatchesCV(e1, e2));
            ret.Expect(Created, new { AsyncLockID = AsyncMonitorID });
            ret.Bag.AsyncLockID = AsyncMonitorID;
            ret.Bag.AsyncConditionVariableID = (ret.Count == 0 || ret[0].Data == null) ? 0 : ret[0].Data.AsyncConditionVariableID;
            return ret;
        }
    }
}
