﻿using System;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using System.Threading.Tasks;
using Nito.AsyncEx;
using System.Linq;
using System.Threading;
using System.Diagnostics.CodeAnalysis;
using Nito.AsyncEx.Internal;

namespace Tests
{
    [ExcludeFromCodeCoverage]
    [TestClass]
    public class AsyncReaderWriterLockUnitTests_NET45
    {
        [TestMethod]
        public async Task Unlocked_PermitsWriterLock()
        {
            var rwl = new AsyncReaderWriterLock();
            await rwl.WriterLockAsync();
        }

        [TestMethod]
        public async Task Unlocked_PermitsMultipleReaderLocks()
        {
            var rwl = new AsyncReaderWriterLock();
            await rwl.ReaderLockAsync();
            await rwl.ReaderLockAsync();
        }

        [TestMethod]
        public async Task WriteLocked_PreventsAnotherWriterLock()
        {
            var rwl = new AsyncReaderWriterLock();
            await rwl.WriterLockAsync();
            var task = rwl.WriterLockAsync();
            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task WriteLocked_PreventsReaderLock()
        {
            var rwl = new AsyncReaderWriterLock();
            await rwl.WriterLockAsync();
            var task = rwl.ReaderLockAsync();
            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task WriteLocked_PreventsUpgradeableReaderLock()
        {
            var rwl = new AsyncReaderWriterLock();
            await rwl.WriterLockAsync();
            var task = rwl.UpgradeableReaderLockAsync();
            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task WriteLocked_Unlocked_PermitsAnotherWriterLock()
        {
            var rwl = new AsyncReaderWriterLock();
            var firstWriteLockTaken = new TaskCompletionSource();
            var releaseFirstWriteLock = new TaskCompletionSource();
            var task = Task.Run(async () =>
            {
                using (await rwl.WriterLockAsync())
                {
                    firstWriteLockTaken.SetResult();
                    await releaseFirstWriteLock.Task;
                }
            });
            await firstWriteLockTaken.Task;
            var lockTask = rwl.WriterLockAsync();
            Assert.IsFalse(lockTask.IsCompleted);
            releaseFirstWriteLock.SetResult();
            await lockTask;
        }

        [TestMethod]
        public async Task ReadLocked_PreventsWriterLock()
        {
            var rwl = new AsyncReaderWriterLock();
            await rwl.ReaderLockAsync();
            var task = rwl.WriterLockAsync();
            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task ReadLocked_AllowsUpgradeableReaderLock()
        {
            var rwl = new AsyncReaderWriterLock();
            await rwl.ReaderLockAsync();
            await rwl.UpgradeableReaderLockAsync();
        }

        [TestMethod]
        public async Task UpgradeableReadLocked_PreventsWriterLock()
        {
            var rwl = new AsyncReaderWriterLock();
            var key = await rwl.UpgradeableReaderLockAsync();
            Assert.IsFalse(key.Upgraded);
            var task = rwl.WriterLockAsync();
            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task UpgradeableReadLocked_AllowsMultipleReaderLocks()
        {
            var rwl = new AsyncReaderWriterLock();
            var key = await rwl.UpgradeableReaderLockAsync();
            Assert.IsFalse(key.Upgraded);
            await rwl.ReaderLockAsync();
            await rwl.ReaderLockAsync();
        }

        [TestMethod]
        public async Task UpgradeableReadLocked_Unlocked_AllowsWriterLock()
        {
            var rwl = new AsyncReaderWriterLock();
            using (var key = await rwl.UpgradeableReaderLockAsync())
            {
                Assert.IsFalse(key.Upgraded);
            }
            await rwl.WriterLockAsync();
        }

        [TestMethod]
        public async Task UpgradeableReadLocked_Upgraded_PreventsReaderLock()
        {
            var rwl = new AsyncReaderWriterLock();
            var key = await rwl.UpgradeableReaderLockAsync();
            await key.UpgradeAsync();
            Assert.IsTrue(key.Upgraded);
            var task = rwl.WriterLockAsync();
            await AssertEx.NeverCompletesAsync(task);
        }

        [TestMethod]
        public async Task UpgradeableReadAndReadLocked_Upgrade_WaitsForReaderLockToUnlock()
        {
            var rwl = new AsyncReaderWriterLock();
            var readLockTaken = new TaskCompletionSource();
            var releaseReadLock = new TaskCompletionSource();
            var task = Task.Run(async () =>
            {
                using (await rwl.ReaderLockAsync())
                {
                    readLockTaken.SetResult();
                    await releaseReadLock.Task;
                }
            });
            await readLockTaken.Task;
            using (var key = await rwl.UpgradeableReaderLockAsync())
            {
                Assert.IsFalse(key.Upgraded);
                var lockTask = key.UpgradeAsync();
                Assert.IsFalse(lockTask.IsCompleted);
                releaseReadLock.SetResult();
                await lockTask;
                Assert.IsTrue(key.Upgraded);
            }
        }

        [TestMethod]
        public async Task UpgradeableReadAndReadLocked_UpgradeAborted_CompletesAsCanceled()
        {
            var rwl = new AsyncReaderWriterLock();
            var readLockTaken = new TaskCompletionSource();
            var task = Task.Run(async () =>
            {
                using (await rwl.ReaderLockAsync())
                {
                    readLockTaken.SetResult();
                    await TaskConstants.Never;
                }
            });
            await readLockTaken.Task;
            Task upgradeTask;
            using (var key = await rwl.UpgradeableReaderLockAsync())
            {
                upgradeTask = key.UpgradeAsync();
                Assert.IsFalse(key.Upgraded);
            }

            await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(upgradeTask);
        }

        [TestMethod]
        public async Task UpgradeableReadLocked_Downgrade_AllowsReaderLock()
        {
            var rwl = new AsyncReaderWriterLock();
            var key = await rwl.UpgradeableReaderLockAsync();
            var upgradeKey = await key.UpgradeAsync();
            upgradeKey.Dispose();
            await rwl.ReaderLockAsync();
        }

        [TestMethod]
        public void Id_IsNotZero()
        {
            var rwl = new AsyncReaderWriterLock();
            Assert.AreNotEqual(0, rwl.Id);
        }

        [TestMethod]
        public void WriterLock_PreCancelled_LockAvailable_SynchronouslyTakesLock()
        {
            var rwl = new AsyncReaderWriterLock();
            var token = new CancellationToken(true);
            
            var task = rwl.WriterLockAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsFalse(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
        }

        [TestMethod]
        public void WriterLock_PreCancelled_LockNotAvailable_SynchronouslyCancels()
        {
            var rwl = new AsyncReaderWriterLock();
            var token = new CancellationToken(true);
            rwl.WriterLockAsync();

            var task = rwl.WriterLockAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsTrue(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
        }

        [TestMethod]
        public void ReaderLock_PreCancelled_LockAvailable_SynchronouslyTakesLock()
        {
            var rwl = new AsyncReaderWriterLock();
            var token = new CancellationToken(true);

            var task = rwl.ReaderLockAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsFalse(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
        }

        [TestMethod]
        public void ReaderLock_PreCancelled_LockNotAvailable_SynchronouslyCancels()
        {
            var rwl = new AsyncReaderWriterLock();
            var token = new CancellationToken(true);
            rwl.WriterLockAsync();

            var task = rwl.ReaderLockAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsTrue(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
        }

        [TestMethod]
        public void UpgradeableReaderLock_PreCancelled_LockAvailable_SynchronouslyTakesLock()
        {
            var rwl = new AsyncReaderWriterLock();
            var token = new CancellationToken(true);

            var task = rwl.UpgradeableReaderLockAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsFalse(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
        }

        [TestMethod]
        public void UpgradeableReaderLock_PreCancelled_LockNotAvailable_SynchronouslyCancels()
        {
            var rwl = new AsyncReaderWriterLock();
            var token = new CancellationToken(true);
            rwl.WriterLockAsync();

            var task = rwl.UpgradeableReaderLockAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsTrue(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
        }

        [TestMethod]
        public void Upgrade_PreCancelled_LockAvailable_SynchronouslyTakesLock()
        {
            var rwl = new AsyncReaderWriterLock();
            var token = new CancellationToken(true);
            var key = rwl.UpgradeableReaderLockAsync().Result;

            var task = key.UpgradeAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsFalse(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
        }

        [TestMethod]
        public void Upgrade_PreCancelled_LockNotAvailable_SynchronouslyCancels()
        {
            var rwl = new AsyncReaderWriterLock();
            var token = new CancellationToken(true);
            var key = rwl.UpgradeableReaderLockAsync().Result;
            rwl.ReaderLockAsync();

            var task = key.UpgradeAsync(token);

            Assert.IsTrue(task.IsCompleted);
            Assert.IsTrue(task.IsCanceled);
            Assert.IsFalse(task.IsFaulted);
        }

        [TestMethod]
        public async Task WriteLocked_WriterLockCancelled_DoesNotTakeLockWhenUnlocked()
        {
            var rwl = new AsyncReaderWriterLock();
            using (await rwl.WriterLockAsync())
            {
                var cts = new CancellationTokenSource();
                var task = rwl.WriterLockAsync(cts.Token);
                cts.Cancel();
                await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
            }

            await rwl.WriterLockAsync();
        }

        [TestMethod]
        public async Task WriteLocked_ReaderLockCancelled_DoesNotTakeLockWhenUnlocked()
        {
            var rwl = new AsyncReaderWriterLock();
            using (await rwl.WriterLockAsync())
            {
                var cts = new CancellationTokenSource();
                var task = rwl.ReaderLockAsync(cts.Token);
                cts.Cancel();
                await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
            }

            await rwl.ReaderLockAsync();
        }

        [TestMethod]
        public async Task WriteLocked_UpgradeableReaderLockCancelled_DoesNotTakeLockWhenUnlocked()
        {
            var rwl = new AsyncReaderWriterLock();
            using (await rwl.WriterLockAsync())
            {
                var cts = new CancellationTokenSource();
                var task = rwl.UpgradeableReaderLockAsync(cts.Token);
                cts.Cancel();
                await AssertEx.ThrowsExceptionAsync<OperationCanceledException>(() => task);
            }

            await rwl.UpgradeableReaderLockAsync();
        }

        [TestMethod]
        public async Task LockReleased_WriteTakesPriorityOverUpgradeableRead()
        {
            var rwl = new AsyncReaderWriterLock();
            Task writeLock, upgradeableReadLock;
            using (await rwl.WriterLockAsync())
            {
                upgradeableReadLock = rwl.UpgradeableReaderLockAsync();
                writeLock = rwl.WriterLockAsync();
            }

            await writeLock;
            await AssertEx.NeverCompletesAsync(upgradeableReadLock);
        }

        [TestMethod]
        public async Task LockReleased_WriteTakesPriorityOverRead()
        {
            var rwl = new AsyncReaderWriterLock();
            Task writeLock, readLock;
            using (await rwl.WriterLockAsync())
            {
                readLock = rwl.ReaderLockAsync();
                writeLock = rwl.WriterLockAsync();
            }

            await writeLock;
            await AssertEx.NeverCompletesAsync(readLock);
        }

        [TestMethod]
        public async Task LockReleased_AllowsUpgradeableReadAndMultipleReaders()
        {
            var rwl = new AsyncReaderWriterLock();
            Task upgradeableReadLock, readLock1, readLock2;
            using (await rwl.WriterLockAsync())
            {
                upgradeableReadLock = rwl.UpgradeableReaderLockAsync();
                readLock1 = rwl.ReaderLockAsync();
                readLock2 = rwl.ReaderLockAsync();
            }

            await Task.WhenAll(upgradeableReadLock, readLock1, readLock2);
        }

        [TestMethod]
        public async Task ReaderLocked_ReaderReleased_ReaderAndWriterWaiting_DoesNotReleaseReaderOrWriter()
        {
            var rwl = new AsyncReaderWriterLock();
            Task readLock, writeLock;
            await rwl.ReaderLockAsync();
            using (await rwl.ReaderLockAsync())
            {
                writeLock = rwl.WriterLockAsync();
                readLock = rwl.ReaderLockAsync();
            }

            await Task.WhenAll(AssertEx.NeverCompletesAsync(writeLock),
                AssertEx.NeverCompletesAsync(readLock));
        }

        [TestMethod]
        public async Task ReaderLocked_ReaderReleased_ReaderAndUpgradingReaderWaiting_DoesNotReleaseReaderOrUpgradingReader()
        {
            var rwl = new AsyncReaderWriterLock();
            Task readLock, upgradingReadLock;
            await rwl.ReaderLockAsync();
            using (await rwl.ReaderLockAsync())
            {
                var upgradeableReadLock = await rwl.UpgradeableReaderLockAsync();
                upgradingReadLock = upgradeableReadLock.UpgradeAsync();
                readLock = rwl.ReaderLockAsync();
            }

            await Task.WhenAll(AssertEx.NeverCompletesAsync(upgradingReadLock),
                AssertEx.NeverCompletesAsync(readLock));
        }

        [TestMethod]
        public async Task ReaderLocked_UpgradableReaderReleased_UpgradableReaderWaiting_ReleasesUpgrableReader()
        {
            var rwl = new AsyncReaderWriterLock();
            Task upgradableReadLock;
            using (await rwl.ReaderLockAsync())
            using (await rwl.UpgradeableReaderLockAsync())
            {
                upgradableReadLock = rwl.UpgradeableReaderLockAsync();
                Assert.IsFalse(upgradableReadLock.IsCompleted);
            }

            await upgradableReadLock;
        }

        [TestMethod]
        public async Task ReaderKey_MultiDispose_DoesNothing()
        {
            var rwl = new AsyncReaderWriterLock();
            var key = await rwl.ReaderLockAsync();
            key.Dispose();
            key.Dispose();
            await rwl.ReaderLockAsync();
        }

        [TestMethod]
        public async Task WriterKey_MultiDispose_DoesNothing()
        {
            var rwl = new AsyncReaderWriterLock();
            var key = await rwl.WriterLockAsync();
            key.Dispose();
            key.Dispose();
            await rwl.WriterLockAsync();
        }

        [TestMethod]
        public async Task UpgradableKey_MultiDispose_DoesNothing()
        {
            var rwl = new AsyncReaderWriterLock();
            var key = await rwl.UpgradeableReaderLockAsync();
            key.Dispose();
            key.Dispose();
            await rwl.UpgradeableReaderLockAsync();
        }

        [TestMethod]
        public async Task UpgradeKey_MultiDispose_DoesNothing()
        {
            var rwl = new AsyncReaderWriterLock();
            var upgradeable = await rwl.UpgradeableReaderLockAsync();
            var key = await upgradeable.UpgradeAsync();
            key.Dispose();
            key.Dispose();
            await upgradeable.UpgradeAsync();
        }

        [TestMethod]
        public async Task UpgradeableKey_MultiUpgrade_ThrowsException()
        {
            var rwl = new AsyncReaderWriterLock();
            var key = await rwl.UpgradeableReaderLockAsync();
            await key.UpgradeAsync();
            await AssertEx.ThrowsExceptionAsync<InvalidOperationException>(async () => { await key.UpgradeAsync(); });
        }

        [TestMethod]
        public async Task UpgradeableKey_MultiUpgradeWhenFirstUpgradeIsIncomplete_ThrowsSynchronousException()
        {
            var rwl = new AsyncReaderWriterLock();
            await rwl.ReaderLockAsync();
            var key = await rwl.UpgradeableReaderLockAsync();
            var _ = key.UpgradeAsync();
            AssertEx.ThrowsException<InvalidOperationException>(() => key.UpgradeAsync());
        }
    }
}