﻿using System;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Diagnostics;
using System.Diagnostics.CodeAnalysis;
using System.Diagnostics.Tracing;
using System.Dynamic;
using System.Globalization;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using System.Xml.Linq;

/// <summary>
/// A listener for ETW events, designed to work well with unit tests and/or a debugger.
/// </summary>
[ExcludeFromCodeCoverage]
[DebuggerDisplay("Count = {_events.Count}")]
[DebuggerTypeProxy(typeof(DebugView))]
internal sealed class DebugEtw : EventListener
{
    private readonly Func<Event, bool> _filter;
    private readonly List<Event> _events = new List<Event>();
    private readonly List<Event> _allEvents = new List<Event>();

    /// <summary>
    /// Constructs an ETW listener with the specified filter.
    /// </summary>
    public DebugEtw(Func<Event, bool> filter)
    {
        _filter = filter ?? (e => true);
    }

    protected override void OnEventSourceCreated(EventSource eventSource)
    {
        // Listen to all events.
        EnableEvents(eventSource, EventLevel.LogAlways, (EventKeywords)(-1));
    }

    protected override void OnEventWritten(EventWrittenEventArgs eventData)
    {
        var e = new Event(eventData);
        if (_filter(e))
        {
            lock (_events)
            {
                _events.Add(e);
                _allEvents.Add(e);
            }
        }
    }

    /// <summary>
    /// Starts an ETW listener with an optional event filter.
    /// </summary>
    public static DebugEtw Start(Func<Event, bool> filter = null)
    {
        return new DebugEtw(filter);
    }

    /// <summary>
    /// Takes a snapshot of the events so far, clears out all events, and continues listening.
    /// </summary>
    public EventCollection Snapshot() { return Snapshot(null, null); }

    /// <summary>
    /// Takes a snapshot of the events so far that match a filter, clears out all events, and continues listening.
    /// </summary>
    public EventCollection Snapshot(Func<Event, bool> filter) { return Snapshot(filter, null); }

    /// <summary>
    /// Takes a snapshot of the events so far that match a filter (or are related to events matching a filter), clears out all events, and continues listening.
    /// </summary>
    public EventCollection Snapshot(Func<Event, bool> filter, Func<Event, Event, bool> related)
    {
        filter = filter ?? (e => true);
        Event[] events;
        lock (_events)
        {
            events = _events.ToArray();
            _events.Clear();
        }

        // Which events are included for any reason. Right now this is just the events passing the filter.
        // When this is set to true, also add the event to the queue below.
        // If this is already set to true, the event is either in the queue or has already been processed.
        var filterResults = events.Select(filter).ToArray();

        if (related != null)
        {
            // The queue of events to process for transitive closures.
            var queue = new Queue<int>();
            for (int i = 0; i != filterResults.Length; ++i)
                if (filterResults[i])
                    queue.Enqueue(i);

            while (queue.Count != 0)
            {
                // Get the next event to process.
                int index = queue.Dequeue();

                // Determine which other events are related to this event, and add them to the queue.
                for (int i = 0; i != events.Length; ++i)
                {
                    if (!filterResults[i] && related(events[index], events[i]))
                    {
                        filterResults[i] = true;
                        queue.Enqueue(i);
                    }
                }
            }
        }

        // Apply the filter results.
        var ret = new List<Event>();
        for (int i = 0; i != filterResults.Length; ++i)
            if (filterResults[i])
                ret.Add(events[i]);

        return new EventCollection(ret.AsReadOnly());
    }

    /// <summary>
    /// Stops listening and returns all events.
    /// </summary>
    public EventCollection Stop() { return Stop(null, null); }

    /// <summary>
    /// Stops listening and returns all events that match a filter.
    /// </summary>
    public EventCollection Stop(Func<Event, bool> filter) { return Stop(filter, null); }

    /// <summary>
    /// Stops listening and returns all events that match a filter (or are related to events matching a filter).
    /// </summary>
    public EventCollection Stop(Func<Event, bool> filter, Func<Event, Event, bool> related)
    {
        var ret = Snapshot(filter, related);
        Dispose();
        return ret;
    }

    private sealed class DebugView
    {
        private readonly DebugEtw _etw;

        public DebugView(DebugEtw etw)
        {
            _etw = etw;
        }

        [DebuggerBrowsable(DebuggerBrowsableState.RootHidden)]
        public Event[] Events
        {
            get
            {
                return _etw._events.ToArray();
            }
        }
    }

    /// <summary>
    /// An ETW event instance.
    /// </summary>
    [DebuggerDisplay("Source = {SourceName}, Event = {EventName}, Message = {Message}")]
    [DebuggerTypeProxy(typeof(DebugView))]
    public sealed class Event
    {
        private static readonly ConcurrentDictionary<EventSource, XElement> _manifests = new ConcurrentDictionary<EventSource, XElement>();
        private static readonly ConcurrentDictionary<Tuple<EventSource, int>, XElement> _eventManifests = new ConcurrentDictionary<Tuple<EventSource, int>, XElement>();
        private static readonly ConcurrentDictionary<Tuple<EventSource, int>, ReadOnlyCollection<string>> _parameterNames = new ConcurrentDictionary<Tuple<EventSource, int>, ReadOnlyCollection<string>>();
        private static readonly XNamespace ns = "http://schemas.microsoft.com/win/2004/08/events";

        private static int ParseInt(string s)
        {
            if (s.StartsWith("0x"))
                return int.Parse(s.Substring(2), NumberStyles.AllowHexSpecifier);
            return int.Parse(s);
        }

        public Event(EventWrittenEventArgs args)
        {
            OriginalEvent = args;
            Bag = new ExpandoObject();

            _eventName = new Lazy<string>(() =>
            {
                // Pull a "friendly name" out of the @template of this event's manifest.
                // This is depending on an implementation detail of EventSource.GenerateManifest but we'll fall back to just the id if anything goes wrong.
                var eventManifest = EventManifest;
                var argsAttribute = eventManifest.Attribute("template");
                var argsName = argsAttribute == null ? null : argsAttribute.Value;
                return argsName != null && argsName.EndsWith("Args") ? argsName.Substring(0, argsName.Length - 4) : Id.ToString();
            });

            _message = new Lazy<string>(() =>
            {
                if (args.Message != null)
                    return string.Format(args.Message, DataValues.ToArray());

                // Try to format the message manually using the manifest.
                // This code may not work in all situations... (MSDN has insufficient documentation)
                var eventManifest = EventManifest;
                var messageAttribute = eventManifest.Attribute("message");
                if (messageAttribute != null && messageAttribute.Value.StartsWith("$(string.") && messageAttribute.Value.EndsWith(")"))
                {
                    var messageId = messageAttribute.Value.Substring(9, messageAttribute.Value.Length - 10);
                    var stringElement = SourceManifest.Descendants(ns + "string").FirstOrDefault(x => x.Attribute("id") != null && x.Attribute("id").Value == messageId);
                    if (stringElement != null && stringElement.Attribute("value") != null)
                    {
                        var format = stringElement.Attribute("value").Value;
                        
                        // Yeah, this will certainly not work for advanced formatting...
                        var stringFormat = Regex.Replace(format, @"%\d+", m =>
                        {
                            var match = m.ToString();
                            int index = int.Parse(match.Substring(1));
                            return "{" + (index - 1) + "}";
                        });

                        return string.Format(stringFormat, DataValues.ToArray());
                    }
                }

                return string.Join("; ", DataValues);
            });

            _data = new Lazy<IDictionary<string, object>>(() =>
            {
                var ret = new ExpandoObject() as IDictionary<string, object>;
                for (int i = 0; i != DataNames.Count; ++i)
                    ret.Add(DataNames[i], DataValues[i]);
                return ret;
            });

            // The debugger doesn't like to display Lazy values, so we make them immediate if the debugger may be looking.
            if (Debugger.IsAttached)
            {
                var _ = _eventName.Value;
                var __ = _message.Value;
                var ___ = _data.Value;
            }
        }

        /// <summary>
        /// Gets the manifest for the event source that raised this event.
        /// </summary>
        public XElement SourceManifest
        {
            get
            {
                return _manifests.GetOrAdd(OriginalEvent.EventSource, s => XElement.Parse(EventSource.GenerateManifest(s.GetType(), string.Empty)));
            }
        }

        /// <summary>
        /// Gets the event-specific portion of the manifest for the event source that raised this event.
        /// </summary>
        public XElement EventManifest
        {
            get
            {
                return _eventManifests.GetOrAdd(Tuple.Create(OriginalEvent.EventSource, Id), k =>
                    SourceManifest.Descendants(ns + "event").First(x => ParseInt(x.Attribute("value").Value) == k.Item2)
                );
            }
        }

        /// <summary>
        /// Gets the names of the event data values. See <see cref="DataValues"/> for the data values, or <see cref="Data"/> for a dynamic representation of the data values.
        /// </summary>
        public ReadOnlyCollection<string> DataNames
        {
            get
            {
                return _parameterNames.GetOrAdd(Tuple.Create(OriginalEvent.EventSource, Id), k =>
                {
                    var manifest = SourceManifest;
                    var eventManifest = EventManifest;
                    var template = manifest.Descendants(ns + "template").First(x => x.Attribute("tid").Value == eventManifest.Attribute("template").Value);
                    var templateData = template.Elements(ns + "data").Select(x => x.Attribute("name").Value);
                    return templateData.ToList().AsReadOnly();
                });
            }
        }

        /// <summary>
        /// Gets the original event with all its raw details.
        /// </summary>
        public EventWrittenEventArgs OriginalEvent { get; private set; }

        /// <summary>
        /// Gets the name of the event source.
        /// </summary>
        public string SourceName { get { return OriginalEvent.EventSource.Name; } }

        /// <summary>
        /// Gets the id of this event definition.
        /// </summary>
        public int Id { get { return OriginalEvent.EventId; } }

        /// <summary>
        /// Gets the version of this event definition.
        /// </summary>
        public byte Version { get { return OriginalEvent.Version; } }

        /// <summary>
        /// Gets the friendly name of this event definition.
        /// </summary>
        public string EventName { get { return _eventName.Value; } }
        private readonly Lazy<string> _eventName;

        /// <summary>
        /// Gets a friendly message string. Note that the message string usually does not contain all the data values.
        /// </summary>
        public string Message { get { return _message.Value; } }
        private readonly Lazy<string> _message;

        /// <summary>
        /// Gets the event data values. See <see cref="DataNames"/> for the names of these values, or <see cref="Data"/> for a dynamic representation of the data values.
        /// </summary>
        public ReadOnlyCollection<object> DataValues { get { return OriginalEvent.Payload; } }

        /// <summary>
        /// Gets a dynamic representation of the event data values.
        /// </summary>
        public dynamic Data { get { return _data.Value; } }
        private readonly Lazy<IDictionary<string, object>> _data;

        /// <summary>
        /// Gets an expando object that can be used to attach data to this event instance.
        /// </summary>
        public dynamic Bag { get; private set; }

        private sealed class DebugView
        {
            private readonly Event _event;

            public DebugView(Event e)
            {
                _event = e;
            }

            [DebuggerBrowsable(DebuggerBrowsableState.RootHidden)]
            public Dictionary<string, object> Data { get { return new Dictionary<string, object>(_event._data.Value); } }
        }
    }

    /// <summary>
    /// A collection of events that were received.
    /// </summary>
    [DebuggerDisplay("Count = {_events.Count}")]
    [DebuggerTypeProxy(typeof(DebugView))]
    public sealed class EventCollection : IReadOnlyList<Event>
    {
        private readonly ReadOnlyCollection<Event> _events;

        public EventCollection(ReadOnlyCollection<Event> events)
        {
            _events = events;
            Bag = new ExpandoObject();
        }

        /// <summary>
        /// Gets an expando object that can be used to attach data to this event collection.
        /// </summary>
        public dynamic Bag { get; private set; }

        public Event this[int index]
        {
            get { return _events[index]; }
        }

        public int Count
        {
            get { return _events.Count; }
        }

        IEnumerator<Event> IEnumerable<Event>.GetEnumerator()
        {
            return _events.GetEnumerator();
        }

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return _events.GetEnumerator();
        }

        private sealed class DebugView
        {
            private readonly EventCollection _events;

            public DebugView(EventCollection events)
            {
                _events = events;
            }

            [DebuggerBrowsable(DebuggerBrowsableState.RootHidden)]
            public Event[] Data { get { return _events._events.ToArray(); } }
        }
    }
}
